from __future__ import print_function

import wx
import wx.grid as gr
from ifigure.utils.edit_list import TextCtrlCopyPaste
from petram.helper.init_helper import *


class InitSettingPanel(wx.Panel):
    def __init__(self, parent, id, setting=None):
        wx.Panel.__init__(self, parent, id)

        sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(sizer)

        sizer2 = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(sizer2, 1, wx.EXPAND | wx.ALL, 5)

        rb1 = wx.RadioButton(self, wx.ID_ANY, "Zero",
                             style=wx.RB_GROUP)
        rb2 = wx.RadioButton(self, wx.ID_ANY, "Use Value ")
        rb3 = wx.RadioButton(self, wx.ID_ANY, "Use init panel value")
        rb4 = wx.RadioButton(self, wx.ID_ANY, "From File")
        rb5 = wx.RadioButton(self, wx.ID_ANY, "From Previous SolveStep")
        rb6 = wx.RadioButton(self, wx.ID_ANY, "Call DWC")

        self.rbs = [rb1, rb2, rb3, rb4, rb5, rb6]

        self.st1 = wx.StaticText(self, wx.ID_ANY, '     value(*): ')
        self.tc1 = TextCtrlCopyPaste(self, wx.ID_ANY, '0.0',
                                     validator=setting['validator'],
                                     style=wx.TE_PROCESS_ENTER)
        self.validator = setting['validator']
        self.st4 = wx.StaticText(self, wx.ID_ANY, '     path:  ')
        self.tc4 = TextCtrlCopyPaste(self, wx.ID_ANY, '',
                                     style=wx.TE_PROCESS_ENTER)
        self.bt4 = wx.Button(self, label='Browse...', style=wx.BU_EXACTFIT)

        self.st6_1 = wx.StaticText(self, wx.ID_ANY, '     dwc:  ')
        self.st6_2 = wx.StaticText(self, wx.ID_ANY, '     args: ')
        self.tc6_1 = TextCtrlCopyPaste(self, wx.ID_ANY, '',
                                       style=wx.TE_PROCESS_ENTER)
        self.tc6_2 = TextCtrlCopyPaste(self, wx.ID_ANY, '',
                                       style=wx.TE_PROCESS_ENTER)

        sizer2.Add(rb1, 0, wx.EXPAND)
        sizer2.Add(rb2, 0, wx.EXPAND)
        sizer3 = wx.BoxSizer(wx.HORIZONTAL)
        sizer3.Add(self.st1, 0, wx.ALL, 2)
        sizer3.Add(self.tc1, 1, wx.EXPAND | wx.ALL, 2)
        sizer2.Add(sizer3, 0, wx.EXPAND)
        sizer2.Add(rb3, 0, wx.EXPAND)
        sizer2.Add(rb4, 0, wx.EXPAND)
        sizer3 = wx.BoxSizer(wx.HORIZONTAL)
        sizer3.Add(self.st4, 0, wx.ALL, 2)
        sizer3.Add(self.tc4, 1, wx.EXPAND | wx.ALL, 2)
        sizer3.Add(self.bt4, 0, wx.ALL, 2)
        sizer2.Add(sizer3, 0, wx.EXPAND)
        sizer2.Add(rb5, 0, wx.EXPAND)
        sizer2.Add(rb6, 0, wx.EXPAND)
        sizer3 = wx.BoxSizer(wx.HORIZONTAL)
        sizer3.Add(self.st6_1, 0, wx.ALL, 2)
        sizer3.Add(self.tc6_1, 1, wx.EXPAND | wx.ALL, 2)
        sizer2.Add(sizer3, 0, wx.EXPAND)
        sizer3 = wx.BoxSizer(wx.HORIZONTAL)
        sizer3.Add(self.st6_2, 0, wx.ALL, 2)
        sizer3.Add(self.tc6_2, 1, wx.EXPAND | wx.ALL, 2)
        sizer2.Add(sizer3, 0, wx.EXPAND)

        self.bt4.Bind(wx.EVT_BUTTON, self.onBrowse)
        for rb in self.rbs:
            self.Bind(wx.EVT_RADIOBUTTON, self.onHit, rb)
        self.Bind(wx.EVT_TEXT_ENTER, self.textctrl_enter, self.tc1)
        self.Bind(wx.EVT_TEXT_ENTER, self.textctrl_enter, self.tc4)
        self.Bind(wx.EVT_TEXT_ENTER, self.textctrl_enter, self.tc6_1)
        self.Bind(wx.EVT_TEXT_ENTER, self.textctrl_enter, self.tc6_2)
        self.rbs[0].SetValue(True)
        self.adjust_enables(0)
        self.Layout()

    def SetValue(self, value):
        for sel, rb in enumerate(self.rbs):
            if sel == value[0]:
                rb.SetValue(True)
            else:
                rb.SetValue(False)
        self.tc1.SetValue(value[1])
        self.tc4.SetValue(value[2])
        self.adjust_enables(value[0])
        self.tc6_1.SetValue(value[3][0])
        self.tc6_2.SetValue(value[3][1])

    def GetValue(self):
        for sel, rb in enumerate(self.rbs):
            if rb.GetValue():
                break
        val = self.tc1.GetValue()
        path = self.tc4.GetValue()
        dwc_name = self.tc6_1.GetValue()
        dwc_params = self.tc6_2.GetValue()
        #if not self.validator(val, None, None): val = '0.0'

        return sel, val, path, (dwc_name, dwc_params)

    def textctrl_enter(self, evt):
        self.send_event(self, evt)

    def Enable(self, value):
        pass

    def onBrowse(self, evt):
        diag = wx.DirDialog(self)
        ret = diag.ShowModal()
        if ret == wx.ID_OK:
            path = diag.GetPath()
            self.tc4.SetValue(path)
            self.send_event(self, evt)
        diag.Destroy()
        evt.Skip()

    def adjust_enables(self, selection):
        if selection == 3:
            self.st4.Enable(True)
            self.tc4.Enable(True)
            self.bt4.Enable(True)
        else:
            self.st4.Enable(False)
            self.tc4.Enable(False)
            self.bt4.Enable(False)
        if selection == 1:
            self.st1.Enable(True)
            self.tc1.Enable(True)
        else:
            self.st1.Enable(False)
            self.tc1.Enable(False)
        if selection == 5:
            self.st6_1.Enable(True)
            self.tc6_1.Enable(True)
            self.st6_2.Enable(True)
            self.tc6_2.Enable(True)
        else:
            self.st6_1.Enable(False)
            self.tc6_1.Enable(False)
            self.st6_2.Enable(False)
            self.tc6_2.Enable(False)

    def onHit(self, evt):
        w = evt.GetEventObject()
        sel = self.rbs.index(w)
        self.adjust_enables(sel)
        self.send_event(self, evt)
        evt.Skip()

    def send_event(self, obj, evt):
        evt.SetEventObject(self)
        if hasattr(self.GetParent(), 'send_event'):
            self.GetParent().send_event(self, evt)


class Example(wx.Frame):
    def __init__(self, parent=None, title='test'):
        super(Example, self).__init__(parent, title=title)
        self.w = InitSettingPanel(self, wx.ID_ANY)
        self.SetSizer(wx.BoxSizer(wx.VERTICAL))
        self.GetSizer().Add(self.w, 1, wx.EXPAND, 0)

        #value = [True, {'k':[('x', 'a*3'), ('y', 'b+b')], 'alpha':[]}]
        self.w.SetValue((1, '500', '/home/shiraiwa', ('dwc', 'test')))

        self.Show()
        self.Bind(wx.EVT_CLOSE, self.onClose)

    def onClose(self, evt):
        print(self.w.GetValue())
        evt.Skip()


if __name__ == '__main__':
    app = wx.App(False)
    e = Example()
    app.MainLoop()



import petram.debug
from petram.mfem_config import use_parallel
from petram.namespace_mixin import NS_mixin
from petram.model import Model
import os
import numpy as np
import mfem
from abc import abstractmethod

PyMFEM_PATH = os.path.dirname(os.path.dirname(mfem.__file__))
PetraM_PATH = os.getenv("PetraM")
HOME = os.path.expanduser("~")


if use_parallel:
    import mfem.par as mfem
    from mpi4py import MPI
    num_proc = MPI.COMM_WORLD.size
    myid = MPI.COMM_WORLD.rank
    from petram.helper.mpi_recipes import *
else:
    import mfem.ser as mfem

dprint1, dprint2, dprint3 = petram.debug.init_dprints('MeshModel')


class Mesh(Model, NS_mixin):
    isMeshGenerator = False
    isRefinement = False         # refinement performed in either serial/parallel
    isSerialRefinement = False   # refinement performed in serial

    def __init__(self, *args, **kwargs):
        super(Mesh, self).__init__(*args, **kwargs)
        NS_mixin.__init__(self, *args, **kwargs)

    def onItemSelChanged(self, evt):
        '''
        GUI response when model object is selected in
        the dlg_edit_model
        '''
        viewer = evt.GetEventObject().GetTopLevelParent().GetParent()
        viewer.set_view_mode('phys', self)

    def get_mesh_root(self):
        from petram.mfem_model import MFEM_MeshRoot

        p = self.parent
        while p is not None:
            if isinstance(p, MFEM_MeshRoot):
                return p
            p = p.parent


class MeshGenerator(Mesh):
    isMeshGenerator = True
    isRefinement = False
    isSerialRefinement = False   # refinement performed in serial

    def attribute_set(self, v):
        v = super(MeshGenerator, self).attribute_set(v)
        v['enforce_ncmesh'] = False
        return v

    def panel1_param(self):
        panels = super(MeshGenerator, self).panel1_param()

        p1 = [None, False, 3, {"text": "EnforceNCMesh"}]

        panels.append(p1)
        return panels

    def get_panel1_value(self):
        values = super(MeshGenerator, self).get_panel1_value()
        values.append(self.enforce_ncmesh)
        return values

    def import_panel1_value(self, v):
        super(MeshGenerator, self).import_panel1_value(v[:-1])
        self.enforce_ncmesh = v[-1]

    def run_serial(self, mesh=None):
        # By default this will call run. Sub-classes can re-implement this.
        m = self.run(mesh=mesh)
        return m

    @abstractmethod
    def run(self, mesh=None):
        pass


class MFEMMesh(Model):
    can_delete = True
    has_2nd_panel = False
    isMeshGroup = True

    def get_possible_child(self):
        try:
            from petram.mesh.pumimesh_model import PumiMesh
            return [MeshFile, PumiMesh, Mesh1D, Mesh2D, Mesh3D,
                    UniformRefinement, DomainRefinement, BoundaryRefinement, Scale]
        except BaseException:
            return [MeshFile, Mesh1D, Mesh2D, Mesh3D, UniformRefinement,
                    DomainRefinement, BoundaryRefinement, Scale]

    def get_possible_child_menu(self):
        try:
            from petram.mesh.pumimesh_model import PumiMesh
            return [("", MeshFile),
                    ("Other Meshes", Mesh1D),
                    ("", Mesh2D),
                    ("", Mesh3D),
                    ("!", PumiMesh),
                    ("", Scale),
                    ("Refinement...", UniformRefinement),
                    ("", DomainRefinement),
                    ("!", BoundaryRefinement)]
        except BaseException:
            return [("", MeshFile),
                    ("Other Meshes", Mesh1D),
                    ("", Mesh2D),
                    ("!", Mesh3D),
                    ("", Scale),
                    ("Refinement...", UniformRefinement),
                    ("", DomainRefinement),
                    ("!", BoundaryRefinement)]

    def panel1_param(self):
        if not hasattr(self, "_topo_check_char"):
            self._topo_check_char = "\n".join([' '*15, ' '*15, ' '*15, ' '*15])
            self._invalid_data = None
        import wx
        return [[None, None, 341, {"label": "Reload mesh",
                                   "func": 'call_reload_mfem_mesh',
                                   "noexpand": True}],
                [None, None, 341, {"label": "Check topology",
                                   "func": 'check_topology',
                                   "noexpand": True}],
                [None, self._topo_check_char, 2, None], ]

    def get_panel1_value(self):
        if not hasattr(self, "_topo_check_char"):
            self._topo_check_char = ''
            self._invalid_data = None

        return [self, self, self._topo_check_char]

    def import_panel1_value(self, v):
        pass

    def onItemSelChanged(self, evt):
        '''
        GUI response when model object is selected in
        the dlg_edit_model
        '''
        viewer = evt.GetEventObject().GetTopLevelParent().GetParent()
        viewer.set_view_mode('phys', self)

    def is_viewmode_grouphead(self):
        return True

    def figure_data_name(self):
        return 'mfem'

    def get_special_menu(self, evt):
        # menu =[["Reload Mesh", self.reload_mfem_mesh, None,],]
        menu = [["+Mesh parameters...", None, None],
                ["Plot low quality elements", self.plot_lowqualities, None], ]

        if (self._invalid_data is not None and
                len(self._invalid_data[0]) > 0):
            menu.append(["Plot invalid faces", self.plot_invalids, None])
        if (self._invalid_data is not None and
                len(self._invalid_data[2]) > 0):
            menu.append(["Plot inverted elements", self.plot_inverted, None])

        menu.append(["Compute minSJac", self.compute_scaled_jac, None])
        menu.append(["!", None, None])
        return menu

    def reload_mfem_mesh(self, evt):
        evt.GetEventObject().GetParent().onLoadMesh(evt)

    def call_reload_mfem_mesh(self, evt):
        editor = evt.GetEventObject().GetTopLevelParent()
        viewer = editor.GetParent()

        viewer.onLoadMesh(evt)
        self._topo_check_char = ''
        self._invalid_data = None
        editor.import_selected_panel_value()

    def check_topology(self, evt):
        from petram.mesh.mesh_inspect import (find_invalid_topology,
                                              format_error)

        editor = evt.GetEventObject().GetTopLevelParent()
        dlg = editor.show_progress_bar('Checking topology...')

        try:
            viewer = editor.GetParent()
            mesh = viewer.model.variables.getvar('mesh')
            if mesh is None:
                out = 'Mesh is not loaded'
            else:
                invalids, invalid_attrs, inverted, sj_min_max = find_invalid_topology(
                    mesh)
                if sj_min_max[0] < 0:
                    out = "\n".join(["Some elements are inverted",
                                     "min(ScaledJac) = " + str(sj_min_max[0]),
                                     "max(ScaledJac) = " + str(sj_min_max[1]), ])
                elif len(invalids) == 0:
                    out = "\n".join(["No error",
                                     "min(ScaledJac) = " + str(sj_min_max[0]),
                                     "max(ScaledJac) = " + str(sj_min_max[1]), ])
                else:
                    out = format_error(invalids, invalid_attrs, inverted)
                self._invalid_data = invalids, invalid_attrs, inverted, sj_min_max
            self._topo_check_char = out

            import wx
            wx.CallAfter(editor.import_selected_panel_value)
        except BaseException:
            import traceback
            traceback.print_exc()
        finally:
            dlg.Destroy()
            evt.Skip()

    def compute_scaled_jac(self, evt):
        from petram.mesh.mesh_inspect import get_scaled_jacobian

        editor = evt.GetEventObject().GetTopLevelParent()
        dlg = editor.show_progress_bar(
            'Computing minimum scaled Jacobian (minSJac)...')

        try:
            viewer = editor.GetParent()
            mesh = viewer.model.variables.getvar('mesh')
            sj = get_scaled_jacobian(mesh, sd=-1)
            fec = mfem.L2_FECollection(0, mesh.Dimension())
            fes = mfem.FiniteElementSpace(mesh, fec)
            vec = mfem.Vector(sj)
            gf = mfem.GridFunction(fes, vec.GetData())

        except BaseException:
            import traceback
            traceback.print_exc()
        finally:
            dlg.Destroy()
            evt.Skip()

        folder = viewer.model.param.eval('sol')
        cwd = os.getcwd()
        os.chdir(folder.owndir())
        viewer.engine.save_solfile_fespace('minSJac', 0, gf, None)
        os.chdir(cwd)

        # this triggers to reload solfile
        viewer.model.variables.setvar('solfiles', None)

    def plot_invalids(self, evt):
        from petram.mesh.mesh_inspect import plot_faces_containing_elements

        editor = evt.GetEventObject().GetTopLevelParent()
        viewer = editor.GetParent()
        mesh = viewer.model.variables.getvar('mesh')

        invalids = self._invalid_data[0]

        from ifigure.interactive import figure
        from petram.mfem_viewer import setup_figure
        win = figure()
        setup_figure(win)
        win.view('noclip')
        plot_faces_containing_elements(mesh, invalids, refine=10,
                                       win=win)

    def plot_inverted(self, evt):
        from petram.mesh.mesh_inspect import plot_elements

        editor = evt.GetEventObject().GetTopLevelParent()
        viewer = editor.GetParent()
        mesh = viewer.model.variables.getvar('mesh')

        inverted = self._invalid_data[2]

        from ifigure.interactive import figure
        from petram.mfem_viewer import setup_figure
        win = figure()
        setup_figure(win)
        win.view('noclip')
        plot_elements(mesh, inverted, refine=10, win=win)

    def plot_lowqualities(self, evt):
        from petram.mesh.mesh_inspect import plot_elements

        editor = evt.GetEventObject().GetTopLevelParent()
        viewer = editor.GetParent()

        mesh = viewer.model.variables.getvar('mesh')
        if mesh is None:
            return
        dim = mesh.Dimension()
        sdim = mesh.SpaceDimension()

        if sdim != dim:
            return

        import wx
        from ifigure.utils.edit_list import DialogEditList

        l = [["minimum kappa :", str(1000),  0, {'noexpand': True}], ]

        value = DialogEditList(l, parent=editor,
                               title="Enter mimimum kappa...",
                               style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER)
        if not value[0]:
            return
        minkappa = float(value[1][0])

        J = mfem.DenseMatrix(sdim, dim)

        def GetElementJacobian(mesh, i):
            bgeom = mesh.GetElementBaseGeometry(i)
            T = mesh.GetElementTransformation(i)
            T.SetIntPoint(mfem.Geometries.GetCenter(bgeom))
            mfem.Geometries.JacToPerfJac(bgeom, T.Jacobian(), J)

        lowq_elements = []
        for i in range(mesh.GetNE()):
            GetElementJacobian(mesh, i)
            kappa = J.CalcSingularvalue(0) / J.CalcSingularvalue(dim-1)
            if kappa > minkappa:
                lowq_elements.append(i)

        print("number of low Q elements: " + str(len(lowq_elements)))
        from ifigure.interactive import figure
        from petram.mfem_viewer import setup_figure
        win = figure()
        setup_figure(win)
        win.view('noclip')
        plot_elements(mesh, lowq_elements, refine=10, win=win)

    @property
    def sdim(self):
        if not hasattr(self, '_sdim'):
            self._sdim = 1
        return self._sdim

    @sdim.setter
    def sdim(self, value):
        self._sdim = value


MeshGroup = MFEMMesh


def format_mesh_characteristic(mesh):
    h_min = mfem.doublep()
    h_max = mfem.doublep()
    kappa_min = mfem.doublep()
    kappa_max = mfem.doublep()
    Vh = mfem.Vector()
    Vk = mfem.Vector()
    mesh.GetCharacteristics(h_min, h_max, kappa_min, kappa_max, Vh, Vk)
    h_min = h_min.value()
    h_max = h_max.value()
    kappa_min = kappa_min.value()
    kappa_max = kappa_max.value()

    out = ["", "=== Mesh Statistics ==="]
    out.append("Dimension          : " + str(mesh.Dimension()))
    out.append("Space dimension    : " + str(mesh.SpaceDimension()))

    if mesh.Dimension() == 0:
        out.append("Number of vertices : " + str(mesh.GetNV()))
        out.append("Number of elements : " + str(mesh.GetNE()))
        out.append("Number of bdr elem : " + str(mesh.GetNBE()))
    elif mesh.Dimension() == 1:
        out.append("Number of vertices : " + str(mesh.GetNV()))
        out.append("Number of elements : " + str(mesh.GetNE()))
        out.append("Number of bdr elem : " + str(mesh.GetNBE()))
        out.append("h_min              : " + str(h_min))
        out.append("h_max              : " + str(h_max))
    elif mesh.Dimension() == 2:
        out.append("Number of vertices : " + str(mesh.GetNV()))
        out.append("Number of edges    : " + str(mesh.GetNEdges()))
        out.append("Number of elements : " + str(mesh.GetNE()))
        out.append("Number of bdr elem : " + str(mesh.GetNBE()))
        out.append("Euler Number       : " + str(mesh.EulerNumber2D()))
        out.append("h_min              : " + str(h_min))
        out.append("h_max              : " + str(h_max))
        out.append("kappa_min              : " + str(kappa_min))
        out.append("kappa_max              : " + str(kappa_max))
    elif mesh.Dimension() == 3:
        out.append("Number of vertices : " + str(mesh.GetNV()))
        out.append("Number of edges    : " + str(mesh.GetNEdges()))
        out.append("Number of faces    : " + str(mesh.GetNFaces()))
        out.append("Number of elements : " + str(mesh.GetNE()))
        out.append("Number of bdr elem : " + str(mesh.GetNBE()))
        out.append("Euler Number       : " + str(mesh.EulerNumber()))
        out.append("h_min              : " + str(h_min))
        out.append("h_max              : " + str(h_max))
        out.append("kappa_min              : " + str(kappa_min))
        out.append("kappa_max              : " + str(kappa_max))
    return '\n'.join(out)


class MeshFile(MeshGenerator):
    has_2nd_panel = False

    def __init__(self, parent=None, **kwargs):
        self.path = kwargs.pop("path", "")
        self.generate_edges = kwargs.pop("generate_edges", 1)
        self.refine = kwargs.pop("refien", 1)
        self.fix_orientation = kwargs.pop("fix_orientation", True)
        super(MeshFile, self).__init__(parent=parent, **kwargs)

    def __repr__(self):
        try:
            return 'MeshFile(' + self.path + ')'
        except BaseException:
            return 'MeshFile(!!!Error!!!)'

    def attribute_set(self, v):
        v = super(MeshFile, self).attribute_set(v)
        v['path'] = ''
        v['generate_edges'] = 1
        v['refine'] = True
        v['fix_orientation'] = True

        return v

    def panel1_param(self):
        if not hasattr(self, "_mesh_char"):
            self._mesh_char = ''
        wc = "ANY|*|MFEM|*.mesh|GMSH|*.gmsh"
        p1 = [["Path", self.path, 45, {'wildcard': wc}],
              ["",
               "note: ~ and environmental variables are expanded. \n    In addition, {petram}=$PetraM, {mfem}=PyMFEM, \n     {home}=~ ,{model}=project file dir.",
               2,
               None],
              [None, self.generate_edges == 1,
               3, {"text": "Generate edges"}],
              [None, self.refine == 1, 3, {"text": "Refine"}],
              [None, self.fix_orientation, 3, {"text": "FixOrientation"}],
              [None, self._mesh_char, 2, None], ]

        p2 = MeshGenerator.panel1_param(self)
        return p1[:-1] + p2 + p1[-1:]

    def get_panel1_value(self):
        v1 = [self.path, None, self.generate_edges,
              self.refine, self.fix_orientation, None]

        v2 = MeshGenerator.get_panel1_value(self)

        return v1[:-1] + v2 + v1[-1:]

    def import_panel1_value(self, v):
        self.path = str(v[0])
        self.generate_edges = 1 if v[2] else 0
        self.refine = 1 if v[3] else 0
        self.fix_orientation = v[4]

        MeshGenerator.import_panel1_value(self, v[4:-1])

    def use_relative_path(self):
        self._path_bk = self.path

        try:
            self.path = os.path.basename(self.get_real_path())
        except AssertionError as error:
            if error.args[0].startswith("Mesh file does not exist :"):
                pass
        except BaseException:
            raise

    def restore_fullpath(self):
        self.path = self._path_bk
        self._path_bk = ''

    def get_real_path(self):
        path = str(self.path)

        if path == '':
            # if path is empty, file is given by internal mesh generator.
            parent = self.get_mesh_root()
            for key in parent.keys():
                if not parent[key].is_enabled():
                    continue
                if hasattr(parent[key], 'get_meshfile_path'):
                    return parent[key].get_meshfile_path()
        import os
        path = os.path.expanduser(path)
        path = os.path.expandvars(path)

        if path.find('{mfem}') != -1:
            path = path.replace('{mfem}', PyMFEM_PATH)
        if path.find('{petram}') != -1:
            path = path.replace('{petram}', PetraM_PATH)
        if path.find('{home}') != -1:
            path = path.replace('{home}', HOME)
        if path.find('{model}') != -1:
            path = path.replace('{model}', str(self.root().model_path))

        if not os.path.isabs(path):
            dprint2("meshfile relative path mode")
            path1 = os.path.join(os.getcwd(), path)
            dprint2("trying :", path1)
            if not os.path.exists(path1):
                path1 = os.path.join(os.path.dirname(os.getcwd()), path)
                dprint2("trying :", path1)
                if (not os.path.exists(path1) and "__main__" in globals()
                        and hasattr(__main__, '__file__')):
                    from __main__ import __file__ as mainfile
                    path1 = os.path.join(os.path.dirname(
                        os.path.realpath(mainfile)), path)
                    dprint1("trying :", path1)
                if not os.path.exists(path1) and os.getenv(
                        'PetraM_MeshDir') is not None:
                    path1 = os.path.join(os.getenv('PetraM_MeshDir'), path)
                    dprint1("trying :", path1)
            if os.path.exists(path1):
                path = path1
            else:
                assert False, "can not find mesh file from relative path: " + path
        return path

    def run(self, mesh=None):
        path = self.get_real_path()
        if not os.path.exists(path):
            print("mesh file does not exists : " + path + " in " + os.getcwd())
            return None
        args = (path, self.generate_edges, self.refine, self.fix_orientation)

        mesh = mfem.Mesh(*args)

        if self.enforce_ncmesh:
            mesh.EnsureNCMesh()

        self.parent.sdim = mesh.SpaceDimension()
        self._mesh_char = format_mesh_characteristic(mesh)
        try:
            mesh.GetNBE()
            return mesh
        except BaseException:
            return None


class Mesh1D(MeshGenerator):
    has_2nd_panel = False
    unique_child = True

    def attribute_set(self, v):
        v = super(Mesh1D, self).attribute_set(v)
        v['length'] = [1, ]
        v['nsegs'] = [100, ]
        v['length_txt'] = "1"
        v['nsegs_txt'] = "100"
        v['refine'] = 1
        v['fix_orientation'] = True
        v['mesh_x0_txt'] = "0.0"
        v['mesh_x0'] = 0.0
        return v

    def panel1_param(self):
        if not hasattr(self, "_mesh_char"):
            self._mesh_char = ''

        def check_int_array(txt, param, w):
            try:
                val = [int(x) for x in txt.split(',')]
                return True
            except BaseException:
                return False

        def check_float_array(txt, param, w):
            try:
                val = [float(x) for x in txt.split(',')]
                return True
            except BaseException:
                return False

        def check_float(txt, param, w):
            try:
                val = float(txt)
                return True
            except BaseException:
                return False

        p1 = [["Length", self.length_txt, 0, {"validator": check_float_array}],
              ["N segments", self.nsegs_txt, 0, {
                  "validator": check_int_array}],
              ["x0", self.mesh_x0_txt, 0, {"validator": check_float}],
              [None, "Note: use comma separated float/integer for a multisegments mesh", 2, {}],
              [None, self._mesh_char, 2, None], ]

        return p1

    def get_panel1_value(self):
        v1 = [self.length_txt, self.nsegs_txt, self.mesh_x0_txt, None, None]
        return v1

    def import_panel1_value(self, v):
        self.length_txt = str(v[0])
        self.nsegs_txt = str(v[1])
        self.mesh_x0_txt = str(v[2])

    def eval_strings(self):
        g = self._global_ns.copy()
        l = {}

        try:
            self.length = [float(eval(x, g, l))
                           for x in self.length_txt.split(',')]
            self.nsegs = [int(eval(x, g, l))
                          for x in self.nsegs_txt.split(',')]
            self.mesh_x0 = float(eval(self.mesh_x0_txt, g, l))
            return True
        except BaseException:
            import traceback
            traceback.print_exc()

        return False

    def run(self, mesh=None):

        from petram.mesh.make_simplemesh import straight_line_mesh

        success = self.eval_strings()
        assert success, "Conversion error of input parameter"

        mesh = straight_line_mesh(self.length, self.nsegs,
                                  filename='',
                                  refine=self.refine == 1,
                                  fix_orientation=self.fix_orientation,
                                  sdim=1, x0=self.mesh_x0)
        self.parent.sdim = mesh.SpaceDimension()
        self._mesh_char = format_mesh_characteristic(mesh)
        try:
            mesh.GetNBE()
            return mesh
        except BaseException:
            return None


class Mesh2D(MeshGenerator):
    isRefinement = False
    has_2nd_panel = False
    unique_child = True

    def attribute_set(self, v):
        v = super(Mesh2D, self).attribute_set(v)
        v['length'] = [1, ]
        v['nsegs'] = [100, ]
        v['xlength_txt'] = "1"
        v['ylength_txt'] = "1"
        v['xnsegs_txt'] = "30"
        v['ynsegs_txt'] = "20"
        v['refine'] = 1
        v['fix_orientation'] = True
        v['mesh_x0_txt'] = "0.0, 0.0"
        v['mesh_x0'] = (0.0, 0.0, )
        return v

    def panel1_param(self):
        if not hasattr(self, "_mesh_char"):
            self._mesh_char = ''

        def check_int_array(txt, param, w):
            try:
                val = [int(x) for x in txt.split(',')]
                return True
            except BaseException:
                return False

        def check_float_array(txt, param, w):
            try:
                val = [float(x) for x in txt.split(',')]
                return True
            except BaseException:
                return False

        def check_float(txt, param, w):
            try:
                val = float(txt)
                return True
            except BaseException:
                return False

        p1 = [["Length(x)", self.xlength_txt, 0, {"validator": check_float_array}],
              ["N segments(x)", self.xnsegs_txt, 0, {
                  "validator": check_int_array}],
              ["Length(y)", self.ylength_txt, 0, {
                  "validator": check_float_array}],
              ["N segments(y)", self.ynsegs_txt, 0, {
                  "validator": check_int_array}],
              ["x0", self.mesh_x0_txt, 0, {"validator": check_float_array}],
              [None, "Note: use comma separated float/integer for a multisegments mesh", 2, {}],
              [None, self._mesh_char, 2, None], ]

        p2 = MeshGenerator.panel1_param(self)
        return p1[:-2] + p2 + p1[-2:]

    def get_panel1_value(self):
        v1 = [self.xlength_txt, self.xnsegs_txt, self.ylength_txt, self.ynsegs_txt,
              self.mesh_x0_txt, None, None]
        v2 = MeshGenerator.get_panel1_value(self)

        return v1[:-2] + v2 + v1[-2:]

    def import_panel1_value(self, v):
        self.xlength_txt = str(v[0])
        self.xnsegs_txt = str(v[1])
        self.ylength_txt = str(v[2])
        self.ynsegs_txt = str(v[3])
        self.mesh_x0_txt = str(v[4])

        MeshGenerator.import_panel1_value(self, v[4:-2])

    def eval_strings(self):
        g = self._global_ns.copy()
        l = {}

        try:
            self.xlength = [float(eval(x, g, l))
                            for x in self.xlength_txt.split(',')]
            self.xnsegs = [int(eval(x, g, l))
                           for x in self.xnsegs_txt.split(',')]
            self.ylength = [float(eval(x, g, l))
                            for x in self.ylength_txt.split(',')]
            self.ynsegs = [int(eval(x, g, l))
                           for x in self.ynsegs_txt.split(',')]
            self.mesh_x0 = [float(eval(x, g, l))
                            for x in self.mesh_x0_txt.split(',')]
            return True
        except BaseException:
            import traceback
            traceback.print_exc()

        return False

    def run(self, mesh=None):

        from petram.mesh.make_simplemesh import quad_rectangle_mesh

        success = self.eval_strings()
        assert success, "Conversion error of input parameter"

        mesh = quad_rectangle_mesh(self.xlength, self.xnsegs, self.ylength, self.ynsegs,
                                   filename='', refine=self.refine == 1,
                                   fix_orientation=self.fix_orientation,
                                   sdim=2, x0=self.mesh_x0)

        if self.enforce_ncmesh:
            mesh.EnsureNCMesh()
        self.parent.sdim = mesh.SpaceDimension()
        self._mesh_char = format_mesh_characteristic(mesh)

        try:
            mesh.GetNBE()
            return mesh
        except BaseException:
            return None


class Mesh3D(MeshGenerator):
    has_2nd_panel = False
    unique_child = True

    def attribute_set(self, v):
        v = super(Mesh3D, self).attribute_set(v)
        v['length'] = [1, ]
        v['nsegs'] = [100, ]
        v['xlength_txt'] = "1"
        v['ylength_txt'] = "1"
        v['zlength_txt'] = "1"
        v['xnsegs_txt'] = "10"
        v['ynsegs_txt'] = "10"
        v['znsegs_txt'] = "10"
        v['refine'] = 1
        v['fix_orientation'] = True
        v['mesh_x0_txt'] = "0.0, 0.0, 0.0"
        v['mesh_x0'] = (0.0, 0.0, 0.0)
        return v

    def panel1_param(self):
        if not hasattr(self, "_mesh_char"):
            self._mesh_char = ''

        def check_int_array(txt, param, w):
            try:
                val = [int(x) for x in txt.split(',')]
                return True
            except BaseException:
                return False

        def check_float_array(txt, param, w):
            try:
                val = [float(x) for x in txt.split(',')]
                return True
            except BaseException:
                return False

        def check_float(txt, param, w):
            try:
                val = float(txt)
                return True
            except BaseException:
                return False

        p1 = [["Length(x)", self.xlength_txt, 0, {"validator": check_float_array}],
              ["N segments(x)", self.xnsegs_txt, 0, {
                  "validator": check_int_array}],
              ["Length(y)", self.ylength_txt, 0, {
                  "validator": check_float_array}],
              ["N segments(y)", self.ynsegs_txt, 0, {
                  "validator": check_int_array}],
              ["Length(z)", self.zlength_txt, 0, {
                  "validator": check_float_array}],
              ["N segments(z)", self.znsegs_txt, 0, {
                  "validator": check_int_array}],
              ["x0", self.mesh_x0_txt, 0, {"validator": check_float_array}],
              [None, "Note: use comma separated float/integer for a multisegments mesh", 2, {}],
              [None, self._mesh_char, 2, None], ]

        p2 = MeshGenerator.panel1_param(self)

        return p1[:-2] + p2 + p1[-2:]

    def get_panel1_value(self):
        v1 = [self.xlength_txt, self.xnsegs_txt, self.ylength_txt, self.ynsegs_txt,
              self.zlength_txt, self.znsegs_txt, self.mesh_x0_txt, None, None]
        v2 = MeshGenerator.get_panel1_value(self)

        return v1[:-2] + v2 + v1[-2:]

    def import_panel1_value(self, v):
        self.xlength_txt = str(v[0])
        self.xnsegs_txt = str(v[1])
        self.ylength_txt = str(v[2])
        self.ynsegs_txt = str(v[3])
        self.zlength_txt = str(v[4])
        self.znsegs_txt = str(v[5])
        self.mesh_x0_txt = str(v[6])

        MeshGenerator.import_panel1_value(self, v[6:-2])

    def eval_strings(self):
        g = self._global_ns.copy()
        l = {}

        try:
            self.xlength = [float(eval(x, g, l))
                            for x in self.xlength_txt.split(',')]
            self.xnsegs = [int(eval(x, g, l))
                           for x in self.xnsegs_txt.split(',')]
            self.ylength = [float(eval(x, g, l))
                            for x in self.ylength_txt.split(',')]
            self.ynsegs = [int(eval(x, g, l))
                           for x in self.ynsegs_txt.split(',')]
            self.zlength = [float(eval(x, g, l))
                            for x in self.zlength_txt.split(',')]
            self.znsegs = [int(eval(x, g, l))
                           for x in self.znsegs_txt.split(',')]
            self.mesh_x0 = [float(eval(x, g, l))
                            for x in self.mesh_x0_txt.split(',')]
            return True
        except BaseException:
            import traceback
            traceback.print_exc()

        return False

    def run(self, mesh=None):
        from petram.mesh.make_simplemesh import hex_box_mesh

        success = self.eval_strings()
        assert success, "Conversion error of input parameter"

        mesh = hex_box_mesh(self.xlength, self.xnsegs, self.ylength, self.ynsegs, self.zlength, self.znsegs,
                            filename='', refine=self.refine == 1, fix_orientation=self.fix_orientation,
                            sdim=3, x0=self.mesh_x0)

        if self.enforce_ncmesh:
            mesh.EnsureNCMesh()

        self.parent.sdim = mesh.SpaceDimension()
        self._mesh_char = format_mesh_characteristic(mesh)

        try:
            mesh.GetNBE()
            return mesh
        except BaseException:
            return None


class UniformRefinement(Mesh):
    isRefinement = True
    has_2nd_panel = False

    def __init__(self, parent=None, **kwargs):
        self.num_refine = kwargs.pop("num_refine", "0")
        super(UniformRefinement, self).__init__(parent=parent, **kwargs)

    def __repr__(self):
        try:
            return 'MeshUniformRefinement(' + self.num_refine + ')'
        except BaseException:
            return 'MeshUniformRefinement(!!!Error!!!)'

    def attribute_set(self, v):
        v = super(UniformRefinement, self).attribute_set(v)
        v['num_refine'] = '0'
        return v

    def panel1_param(self):
        return [["Number", str(self.num_refine), 0, {}], ]

    def import_panel1_value(self, v):
        self.num_refine = str(v[0])

    def get_panel1_value(self):
        return (str(self.num_refine),)

    def run(self, mesh):
        gtype = np.unique([mesh.GetElementBaseGeometry(i)
                           for i in range(mesh.GetNE())])
        if use_parallel:
            from mpi4py import MPI
            gtype = gtype.astype(np.int32)
            gtype = np.unique(allgather_vector(gtype, MPI.INT))

        if len(gtype) > 1:
            dprint1(
                "(Warning) Element Geometry Type is mixed. Cannot perform UniformRefinement")
            return mesh
        for i in range(int(self.num_refine)):
            mesh.UniformRefinement()  # this is parallel refinement
        return mesh


class Scale(Mesh):
    def attribute_set(self, v):
        v = super(Scale, self).attribute_set(v)
        v['scale'] = '1.0, 1.0, 1.0'
        v['scale_ns'] = 'global'
        return v

    def panel1_param(self):
        return [["scale", self.scale, 0, {}, ],
                ["NS for expr.", self.scale_ns, 0, {}], ]

    def import_panel1_value(self, v):
        self.scale = str(v[0])
        self.scale_ns = str(v[1])

    def get_panel1_value(self):
        return (str(self.scale),
                str(self.scale_ns), )

    def run(self, mesh):
        if self.scale != '':
            code = compile(self.scale, '<string>', 'eval')
            names = list(code.co_names)
            ns_obj, ns_g = self.find_ns_by_name(self.scale_ns)

        ll = {}
        value = eval(code, ns_g, ll)

        sdim = mesh.SpaceDimension()
        # nicePrint("refining elements domain choice", domains)
        for v in mesh.GetVertexArray():
            for i in range(sdim):
                v[i] = v[i]*value[i]

        return mesh


class DomainRefinement(Mesh):
    isRefinement = True
    has_2nd_panel = False

    def __init__(self, parent=None, **kwargs):
        self.num_refine = kwargs.pop("num_refine", "0")
        self.expression = kwargs.pop("expression", "")
        super(DomainRefinement, self).__init__(parent=parent, **kwargs)

    def __repr__(self):
        try:
            return 'MeshUniformRefinement(' + self.num_refine + ')'
        except BaseException:
            return 'MeshUniformRefinement(!!!Error!!!)'

    def attribute_set(self, v):
        v = super(DomainRefinement, self).attribute_set(v)
        v['num_refine'] = '0'
        v['expression'] = ''
        v['expression_ns'] = 'global'
        return v

    def panel1_param(self):
        from petram.model import validate_sel

        return [["Number", str(self.num_refine), 0, {}],
                ["Domains", self.sel_index_txt, 0, {'changing_event': True,
                                                    'setfocus_event': True,
                                                    'validator': validate_sel,
                                                    'validator_param': self}, ],
                ["Expr.", self.expression, 0, {}, ],
                ["NS for expr.", self.expression_ns, 0, {}], ]

    def import_panel1_value(self, v):
        self.num_refine = str(v[0])
        self.sel_index_txt = str(v[1])
        self.expression = str(v[2])
        self.expression_ns = str(v[3])

    def get_panel1_value(self):
        return (str(self.num_refine),
                str(self.sel_index_txt),
                str(self.expression),
                str(self.expression_ns), )

    def run(self, mesh):
        gtype = np.unique([mesh.GetElementBaseGeometry(i)
                           for i in range(mesh.GetNE())])
        if use_parallel:
            from mpi4py import MPI
            gtype = gtype.astype(np.int32)
            gtype = np.unique(allgather_vector(gtype, MPI.INT))

        if len(gtype) > 1:
            dprint1(
                "(Warning) Element Geometry Type is mixed. Cannot perform UniformRefinement")
            return mesh

        domains = self.process_sel_index()
        if len(domains) == 0:
            return mesh

        if self.expression != '':
            code = compile(self.expression, '<string>', 'eval')
            names = list(code.co_names)
            ns_obj, ns_g = self.find_ns_by_name(self.expression_ns)

        v = mfem.Vector()
        coords = ['x', 'y', 'z']
        # nicePrint("refining elements domain choice", domains)
        for i in range(int(self.num_refine)):
            attr = mesh.GetAttributeArray()
            idx = list(np.where(np.in1d(attr, domains))[0])

            if self.expression != '':
                idx2 = []

                for ii in idx:
                    ll = {}
                    vv = mesh.GetElementCenterArray(ii)
                    for i in range(len(vv)):
                        ll[coords[i]] = vv[i]
                    if eval(code, ns_g, ll):
                        idx2.append(ii)

                idx = idx2
            # nicePrint("number of refined element: ", len(idx))
            idx0 = mfem.intArray(idx)
            mesh.GeneralRefinement(idx0)  # this is parallel refinement
        return mesh


class BoundaryRefinement(Mesh):
    isRefinement = True
    isSerialRefinement = True
    has_2nd_panel = False

    def __init__(self, parent=None, **kwargs):
        self.num_refine = kwargs.pop("num_refine", "0")
        self.num_layer = kwargs.pop("num_layer", "4")
        super(BoundaryRefinement, self).__init__(parent=parent, **kwargs)

    def __repr__(self):
        try:
            return 'MeshBoundaryRefinement(' + self.num_refine + ')'
        except BaseException:
            return 'MeshBoundaryRefinement(!!!Error!!!)'

    def attribute_set(self, v):
        v = super(BoundaryRefinement, self).attribute_set(v)
        v['num_refine'] = '0'
        v['num_layer'] = '4'
        return v

    def panel1_param(self):
        from petram.model import validate_sel

        return [["Number", str(self.num_refine), 0, {}],
                ["Boundaries", self.sel_index_txt, 0, {'changing_event': True,
                                                       'setfocus_event': True,
                                                       'validator': validate_sel,
                                                       'validator_param': self}, ],
                ["#Layers", self.num_layer, 0, {}], ]

    def import_panel1_value(self, v):
        self.num_refine = str(v[0])
        self.sel_index_txt = str(v[1])
        self.num_layer = str(v[2])

    def get_panel1_value(self):
        return (str(self.num_refine),
                str(self.sel_index_txt),
                str(self.num_layer),)

    def run(self, mesh):

        from petram.helper.boundary_refinement import apply_boundary_refinement

        nlayers = int(self.num_layer)
        sels = self.process_sel_index()

        ne0 = mesh.GetNE()

        krefine = int(self.num_refine)
        for i in range(krefine):
            mesh = apply_boundary_refinement(mesh, sels, nlayers=nlayers)

        ne1 = mesh.GetNE()
        dprint1("Number of element before/after boundary refinementmesh: " +
                str(ne0) + " -->> " + str(ne1))
        return mesh



import numpy as np
from scipy.sparse import csr_matrix, coo_matrix, lil_matrix

import petram
dprint1, dprint2, dprint3 = petram.debug.init_dprints('dof_mapping_matrix')

from petram.helper.find_dof_map_nd import find_dof_map_nd
from petram.helper.find_dof_map_h1 import find_dof_map_h1

# default mapper
def def_map_u(*args):
    return args[0]   # x -> u
def def_map_v(*args):
    return args[1]   # y -> v

from petram.mfem_config import use_parallel

if use_parallel:
   from petram.helper.mpi_recipes import allgather_vector
   import mfem.par as mfem
else:
   import mfem.ser as mfem
   num_proc = 1
   myid = 0

def dof_mapping_matrix(src,  dst,  fes, tdof, engine=None, dphase=0.0,
                       map_to_u = def_map_u, 
                       map_to_v = def_map_v,
                       smap_to_u = None,
                       smap_to_v = None,
                       tol = 1e-7):
    '''
     map: destinatiom mapping 
     smap: source mapping
    '''
    fec_name = fes.FEColl().Name()

    if fec_name.startswith('ND'):
        mapper = find_dof_map_nd
    elif fec_name.startswith('H1'):
        mapper = find_dof_map_h1
    else:
        raise NotImplementedError("mapping for " + fec_name)

    if smap_to_u is None:
        smap_to_u = map_to_u
    if smap_to_v is None:
        smap_to_v = map_to_v
 


    map = mapper(src,  dst,  map_to_u, map_to_v, smap_to_u, smap_to_v, 
                 fes, engine,  tdof, tol=tol)

    if (dphase == 0.):
        pass
    elif (dphase == 180.):
        map = -map
    else:   
        map = map.astype(complex)        
        map *= np.exp(-1j*np.pi/180*dphase)

    coo = coo_matrix(map)
    r = coo.row
    c = coo.col

    idx = range(map.shape[0])
    map.setdiag(1.0)
    for k in c :  map[k,k] = 0.0

    if use_parallel:
        start_row = fes.GetMyTDofOffset()
        end_row = fes.GetMyTDofOffset() + fes.GetTrueVSize()
        c =  allgather_vector(c)

        map = map.transpose()
        m1 = csr_matrix(map.real[start_row:end_row, :], dtype=float)
        m2 = csr_matrix(map.imag[start_row:end_row, :], dtype=float)
        m1.eliminate_zeros()
        m2.eliminate_zeros()
        from mfem.common.chypre import CHypreMat
        dprint1(m1.shape, m2.shape)
        M = CHypreMat(m1, m2).transpose()

    else:
        from petram.helper.block_matrix import convert_to_ScipyCoo        
        M = convert_to_ScipyCoo(coo_matrix(map, dtype=map.dtype))

    #idx = range(M.shape[0])
    #M.setDiag(idx, 1.0)
    #M.setDiag(c, 0.0)

    return M, r, c




import wx

def progressbar(parent, message, title, count, can_abort=True):
    style = wx.PD_AUTO_HIDE|wx.PD_APP_MODAL
    style = style|wx.PD_CAN_ABORT
    dlg = wx.GenericProgressDialog(title,
                                   message,
                                   count,
                                   parent,
                                   style=style)
    def close_dlg(evt, dlg=dlg):
        dlg.Destroy()
    dlg.Bind(wx.EVT_CLOSE, close_dlg)
    return dlg



'''
  small utility to load mfem with use_parallel
'''
def load(use_parallel):
    if use_parallel:
        import mfem.par as mfem
        from mpi4py import MPI        
    else:
        import mfem.ser as mfem
        MPI= None

    import petram
    import petram.mfem_config
    petram.mfem_config.use_parallel = use_parallel
    
    return mfem, MPI




import os
import wx
from PIL import Image

import logging
logging.getLogger('PIL').setLevel(logging.WARNING)

img_path = os.path.join(os.path.dirname(__file__), '..', 'data', 'icon')
bmp_data = {}


class WidgetForms(wx.Panel):
    def __init__(self, parent, id, *args, **kwargs):
        wx.Panel.__init__(self, parent, id)

        sizer1 = wx.BoxSizer(wx.HORIZONTAL)
        sizer2 = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(sizer1)
        sizer1.Add(sizer2, 1, wx.EXPAND)

        bcb = wx.adv.BitmapComboBox(self, pos=(25, 25), size=(200, -1))

        self.bcb = bcb

        setting = kwargs.pop("setting")
        self.choices_cb = setting['choices_cb']

        choices = ["no integrator available"]
        self._current_choices = ["no integrator available"]
        self._value = "no integrator available"

        self.SetChoices(choices, index=0)

        self.Bind(wx.EVT_COMBOBOX, self.onHit, self.bcb)
        self.Bind(wx.EVT_COMBOBOX_DROPDOWN, self.onDropDown, self.bcb)

        sizer2.Add(self.bcb, 1, wx.EXPAND | wx.ALL, 1)

    def onHit(self, evt):
        sel = self.bcb.GetValue()
        self._value = sel
        self.GetParent().send_event(self, evt)

    def GetValue(self):
        sel = self.bcb.GetValue()
        self._value = sel
        return sel

    def SetValue(self, value):
        self._value = value
        ch = self.choices_cb()
        self.SetChoices(ch)

    def onDropDown(self, evt):
        sel = self._value

        ch = self.choices_cb()
        if sel in ch:
            idx = ch.index(sel)
        else:
            idx = 0
        self.SetChoices(ch, index=idx)

    def SetChoices(self, choices, index=-1):
        sel = self._value

        if len(choices) == 0:
            choices = ["no integrator available"]
            sel = "no integrator available"
            self._value = "no integrator available"

        self.bcb.Clear()

        for x in choices:
            if x == "no integrator available":
                x = "none"

            if not x in bmp_data:
                name = os.path.join(img_path, 'form_' + x + '.png')
                img = Image.open(name)
                size = img.size
                fac = 33/size[1]
                img1 = img.resize((int(size[0]*fac), int(size[1]*fac)),
                                  resample=Image.LANCZOS,)

                image = wx.Image(*img1.size)
                image.SetData(img1.convert("RGB").tobytes())

                bmp = wx.Bitmap(image)
                bmp_data[x] = bmp
            else:
                bmp = bmp_data[x]

            if x == "none":
                self.bcb.Append("no integrator available", bmp, "none")
            else:
                self.bcb.Append(x, bmp, x)

        self._current_choices = choices

        if index != -1:
            self.bcb.SetSelection(index)
        else:
            if sel in choices:
                index = choices.index(sel)
            elif sel == "no integrator available":
                index = 0
            else:
                index = 0
            self.bcb.SetSelection(index)



class Phys(object):
    def __init__(self, mesh = None, domain = None):
        self.mesh = mesh
        self.domain = domain if domain is not None else []
        self.vol = []
        self.face = []
        self.edge = []
        self.point = []
    def __del__(self):
        self.vol = None
        self.face = None
        self.edge = None
        self.point = None
        
    def add_vol(self, vol):
        self.vol.append(vol)
        
    def add_face(self, face):
        self.face.append(face)

    def add_edge(self, edge):
        self.edge.append(edge)

    def add_point(self, point):
        self.point.append(point)

                          
                                       
                         
        



import os
from petram.mesh.mesh_model import Mesh

is_licenses_initialized = False

class PumiMesh(Mesh):
    isMeshGenerator = True   
    isRefinement = False   
    has_2nd_panel = False        
    def __init__(self, parent = None, **kwargs):
        self.path = kwargs.pop("path", "")
        self.generate_edges = kwargs.pop("generate_edges", 1)
        self.refine = kwargs.pop("refien", 1)
        self.fix_orientation = kwargs.pop("fix_orientation", True)        
        super(PumiMesh, self).__init__(parent = parent, **kwargs)

    def __repr__(self):
        try:
           return 'PumiMesh('+self.mesh_path+')'
        except:
           return 'PumiMesh(!!!Error!!!)'
        
    def attribute_set(self, v):
        v = super(PumiMesh, self).attribute_set(v)
        v['mesh_path'] = ''
        v['model_path'] = ''        
        v['generate_edges'] = 1
        v['refine'] = True
        v['fix_orientation'] = True

        return v
        
    def panel1_param(self):
        return [["Model Path",   self.model_path,  200, {}],
                ["Mesh Path",   self.mesh_path,  200, {}],                
                ["", "rule: {petram}=$PetraM, {mfem}=PyMFEM, \n     {home}=~ ,{model}=project file dir."  ,2, None],
                ["Generate edges",    self.generate_edges == 1,  3, {"text":""}],
                ["Refine",    self.refine==1 ,  3, {"text":""}],
                ["FixOrientation",    self.fix_orientation ,  3, {"text":""}]]
    def get_panel1_value(self):
        return (self.mesh_path, self.model_path, None, self.generate_edges, self.refine, self.fix_orientation)
    
    def import_panel1_value(self, v):
        self.mesh_path = str(v[0])
        self.model_path = str(v[1])        
        self.generate_edges = 1 if v[3] else 0
        self.refine = 1 if v[4] else 0
        self.fix_orientation = v[5]
        
    def use_relative_path(self):
        self._path_bk  = self.path
        self.path = os.path.basename(self.get_real_path())

        
    def restore_fullpath(self):       
        self.path = self._path_bk
        self._path_bk = ''


    def get_real_path(self, path):
        if path == '':
           # if path is empty, file is given by internal mesh generator.
           parent = self.get_mesh_root()
           for key in parent.keys():
              if not parent[key].is_enabled(): continue
              if hasattr(parent[key], 'get_meshfile_path'):
                 return parent[key].get_meshfile_path()
        if path.find('{mfem}') != -1:
            path = path.replace('{mfem}', PyMFEM_PATH)
        if path.find('{petram}') != -1:
            path = path.replace('{petram}', PetraM_PATH)
        if path.find('{home}') != -1:
            path = path.replace('{home}', HOME)
        if path.find('{model}') != -1:
            path = path.replace('{model}', str(self.root().model_path))

        if not os.path.isabs(path):
            dprint2("meshfile relative path mode")
            path1 = os.path.join(os.getcwd(), path)
            dprint2("trying :", path1)
            if not os.path.exists(path1):
                path1 = os.path.join(os.path.dirname(os.getcwd()), path)
                dprint2("trying :", path1)
                if (not os.path.exists(path1) and "__main__" in globals() and hasattr(__main__, '__file__')):
                    from __main__ import __file__ as mainfile        
                    path1 = os.path.join(os.path.dirname(os.path.realpath(mainfile)), path)   
                    dprint1("trying :", path1)
                if not os.path.exists(path1) and os.getenv('PetraM_MeshDir') is not None:
                    path1 = os.path.join(os.getenv('PetraM_MeshDir'), path)
                    dprint1("trying :", path1)                    
            if os.path.exists(path1):
                path = path1
            else:
                assert False, "can not find mesh file from relative path: "+path
        return path

    def run(self, mesh = None):
        model_path = self.get_real_path(self.model_path)
        mesh_path = self.get_real_path(self.mesh_path)       

        if not os.path.exists(mesh_path):
            print("mesh file does not exists : " + path + " in " + os.getcwd())
            return None
        if not os.path.exists(model_path):
            print("model file does not exists : " + path + " in " + os.getcwd())
            return None

        ####

        if not globals()['is_licenses_initialized']:
            print("do license etc here ...once")
            globals()['is_licenses_initialized'] = True
            
        assert False, "not implemented : pumi_mesh must return mfem mesh"


        '''
        args = (path,  self.generate_edges, self.refine, self.fix_orientation)
        mesh =  mfem.Mesh(*args)
        self.parent.sdim = mesh.SpaceDimension()
        try:
           mesh.GetNBE()
           return mesh
        except:
           return None
        '''
        



'''
   discrete variable interpolator

   Interface to DiscreteInterpolators in Postprocessing
     


'''
import numpy as np
import traceback
import warnings
import weakref

import petram.debug
dprint1, dprint2, dprint3 = petram.debug.init_dprints('Integration(PP)')

from petram.phys.vtable import VtableElement, Vtable, Vtable_mixin

from petram.postprocess.pp_model import PostProcessBase

from petram.phys.weakform import get_integrators
bilinintegs = get_integrators('BilinearOps')
linintegs = get_integrators('LinearOps')

def add_variable_vector(obj, v, names, gfr, gfi):
    from petram.helper.variables import add_components

    for phys_name in obj.root()['Phys']:
        if not obj.root()['Phys'][phys_name].enabled:
            continue
        if obj.root()['Phys'][phys_name].emesh_idx == gfr._emesh_idx:
            ind_vars = obj.root()['Phys'][phys_name].ind_vars
            break

    ind_vars = [x.strip() for x in ind_vars.split(',') if x.strip() != '']

    add_components(v, names[0], "", ind_vars, gfr, gfi)

def add_variable_scalar(obj, v, names, gfr, gfi):
    from petram.helper.variables import add_scalar

    for phys_name in obj.root()['Phys']:
        if not obj.root()['Phys'][phys_name].enabled:
            continue
        if obj.root()['Phys'][phys_name].emesh_idx == gfr._emesh_idx:
            ind_vars = obj.root()['Phys'][phys_name].ind_vars
            break

    ind_vars = [x.strip() for x in ind_vars.split(',') if x.strip() != '']

    add_scalar(v, names[0], "", ind_vars, gfr, gfi)

class DSInterpolator(PostProcessBase, Vtable_mixin):
    has_2nd_panel = False
    
    @property
    def geom_dim(self):  # dim of geometry
        return self.root()['Mesh'].sdim
    
    def attribute_set(self, v):
        v = super(DSInterpolator, self).attribute_set(v)
        v["interpolation_name"] = 'interp'        
        v['variables'] = ''
        v['sel_index'] = ['all']
        v['sel_index_txt'] = 'all'
        v['order'] = 0        
        return v

    def panel1_param(self):
        ll = [["name", "", 0, {}],
              ["variable", self.variables, 0, {}],
              ["order", self.order, 400, {}],]        
        return ll
    
    def get_panel1_value(self):
        return [self.interpolation_name,
                self.variables,
                self.order]
    
    def import_panel1_value(self, v):
        self.interpolation_name = str(v[0])
        self.variables = str(v[1])
        self.order = int(v[2])

    def panel1_tip(self):
        pass

    def update_dom_selection(self, all_sel=None):
        pass


    def add_variables(self, v, names, gfr, gfi):
        assert False, "need to be implemented in subclass"
        
    def soldict_to_solvars(self, soldict, variables):
        suffix = ""
        names = [self.interpolation_name.strip()]
        fname = self.interpolation_name.strip()

        for k in soldict:
            n = '_'.join(k.split('_')[:-1])
            if n == fname:
               sol = soldict[k]
               solr = sol[0]
               soli = sol[1] if len(sol) > 1 else None
               self.add_variables(variables, names, solr, soli)


class Grad(DSInterpolator):
    @classmethod
    def fancy_menu_name(self):
        return "Grad"

    @classmethod
    def fancy_tree_name(self):
        return 'Grad'

    def attribute_set(self, v):
        v = super(Grad, self).attribute_set(v)
        v['interpolation_name'] = 'grad'        
        return v

    def add_variables(self, v, names, gfr, gfi):
        add_variable_vector(self, v, names, gfr, gfi)

    def run_postprocess(self, engine):
        dprint1("running postprocess: " + self.name())

        name = self.variables.strip()
        if name not in engine.model._variables:
            print(list(engine.model._variables))
            assert False, name + " is not defined"

        var1 = engine.model._variables[name]

        emesh_idx1 = var1.get_emesh_idx()
        emesh_idx = emesh_idx1[0]

        fes1 = engine.fespaces[name]
        element = 'ND_FECollection'

        _is_new, fes2 = engine.get_or_allocate_fecfes(self.interpolation_name,
                                                          emesh_idx,
                                                          element,
                                                          self.order,
                                                          1,)

        gfr = engine.new_gf(fes2)
        if var1.complex:
            gfi = engine.new_gf(fes2)
        else:
            gfi = None

        from petram.helper.operators import Gradient
        opr = Gradient()
        opr._engine = weakref.ref(engine)
        M = opr.assemble(trial=fes1, test=fes2)

        V1 = engine.variable2vector(var1)
        value = M.dot(V1)
        dprint1("operator size", M.shape)

        if var1.complex:
            engine.X2x(value.real, gfr)
            engine.X2x(value.imag, gfi)
        else:
            engine.X2x(value.real, gfr)
            
        from petram.helper.variables import Variables
        
        v = Variables()

        names = [self.interpolation_name.strip()]
        self.add_variables(v, names, gfr, gfi)

        engine.add_PP_to_NS(v)
        engine.save_solfile_fespace(self.interpolation_name, emesh_idx, gfr, gfi)

class Curl(DSInterpolator):
    @classmethod    
    def fancy_menu_name(self):
        return "Curl"
    
    @classmethod    
    def fancy_tree_name(self):
        return 'Curl'
    
    def attribute_set(self, v):
        v = super(Curl, self).attribute_set(v)
        v['interpolation_name'] = 'curl'
        return v

    def add_variables(self, v, names, gfr, gfi):
        fec_name = gfr.FESpace().FEColl().Name()
        if fec_name.startswith('RT'):
            add_variable_vector(self, v, names, gfr, gfi)
        else:
            add_variable_scalar(self, v, names, gfr, gfi)
            
    def run_postprocess(self, engine):
        dprint1("running postprocess: " + self.name())

        name = self.variables.strip()
        if name not in engine.model._variables:
            print(list(engine.model._variables))
            assert False, name + " is not defined"

        var1 = engine.model._variables[name]

        emesh_idx1 = var1.get_emesh_idx()
        emesh_idx = emesh_idx1[0]

        fes1 = engine.fespaces[name]

        dim1 = fes1.GetMesh().Dimension()
        if dim1 == 3:
            element = 'RT_FECollection'
        elif dim1 == 2:
            element = 'L2_FECollection'
        else:
            assert False, "curl supports 2D/3D only"
            
        _is_new, fes2 = engine.get_or_allocate_fecfes(self.interpolation_name,
                                                          emesh_idx,
                                                          element,
                                                          self.order,
                                                          1,)

        gfr = engine.new_gf(fes2)
        if var1.complex:
            gfi = engine.new_gf(fes2)
        else:
            gfi = None

        from petram.helper.operators import Curl
        opr = Curl()
        opr._engine = weakref.ref(engine)
        M = opr.assemble(trial=fes1, test=fes2)

        V1 = engine.variable2vector(var1)
        value = M.dot(V1)

        dprint1("operator size", M.shape)        

        if var1.complex:
            engine.X2x(value.real, gfr)
            engine.X2x(value.imag, gfi)
        else:
            engine.X2x(value.real, gfr)
            
        from petram.helper.variables import Variables
        
        v = Variables()        
        names = [self.interpolation_name.strip()]
        self.add_variables(v, names, gfr, gfi)

        engine.add_PP_to_NS(v)
        engine.save_solfile_fespace(self.interpolation_name, emesh_idx, gfr, gfi)


class Div(DSInterpolator):
    @classmethod    
    def fancy_menu_name(self):
        return "Div"
    
    @classmethod    
    def fancy_tree_name(self):
        return 'Div'
    
    def attribute_set(self, v):
        v = super(Div, self).attribute_set(v)
        v['interpolation_name'] = 'div'                
        return v

    def add_variables(self, v, names, gfr, gfi):
        add_variable_scalar(self, v, names, gfr, gfi)

    def run_postprocess(self, engine):
        dprint1("running postprocess: " + self.name())

        name = self.variables.strip()
        if name not in engine.model._variables:
            print(list(engine.model._variables))
            assert False, name + " is not defined"

        var1 = engine.model._variables[name]

        emesh_idx1 = var1.get_emesh_idx()
        emesh_idx = emesh_idx1[0]

        fes1 = engine.fespaces[name]
        element = 'L2_FECollection'

        _is_new, fes2 = engine.get_or_allocate_fecfes(self.interpolation_name,
                                                          emesh_idx,
                                                          element,
                                                          self.order,
                                                          1,)

        gfr = engine.new_gf(fes2)
        if var1.complex:
            gfi = engine.new_gf(fes2)
        else:
            gfi = None

        from petram.helper.operators import Divergence
        opr = Divergence()
        opr._engine = weakref.ref(engine)
        M = opr.assemble(trial=fes1, test=fes2)

        V1 = engine.variable2vector(var1)
        value = M.dot(V1)

        dprint1("operator size", M.shape)                

        if var1.complex:
            engine.X2x(value.real, gfr)
            engine.X2x(value.imag, gfi)
        else:
            engine.X2x(value.real, gfr)
            
        from petram.helper.variables import Variables
        
        v = Variables()        
        names = [self.interpolation_name.strip()]
        self.add_variables(v, names, gfr, gfi)

        engine.add_PP_to_NS(v)
        engine.save_solfile_fespace(self.interpolation_name, emesh_idx, gfr, gfi)
    



import numpy as np
from scipy.sparse import lil_matrix
import itertools
from collections import defaultdict, OrderedDict
from mfem.common.mpi_debug import nicePrint
from mfem.common.parcsr_extra import ToScipyCoo

from petram.helper.dof_map import get_empty_map

import petram.debug
dprint1, dprint2, dprint3 = petram.debug.init_dprints('convolve')

from petram.mfem_config import use_parallel
if use_parallel:
    USE_PARALLEL = True
    import mfem.par as mfem
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
    myid = comm.rank
    nprc = comm.size
else:
    USE_PARALLEL = False    
    import mfem.ser as mfem
    myid = 0
    nprc = 1
    mfem.ParFiniteElementSpace = type(None)
    mfem.ParGridfunction = type(None)
    mfem.ParMesh = type(None)
    mfem.ParMixedBilinearForm = type(None)

def delta(x, _x0, w=None):
    '''
    delta function like coefficient.
    has to return 1/w instead of 1.
    (this is for debugging)
    '''
    #return 1.0
    if np.sqrt(np.sum(x**2)) < 1e-5:
        return 1/w
    return 0.0

def zero(x):
    '''
    sample support funciton. 0.0 means that the contribution is non-zero
    only when x1 and x2 are the same location
    '''
    return 0.0

def get_rule(fe1, fe2, trans, orderinc, verbose):
    order = fe1.GetOrder() + fe2.GetOrder() + trans.OrderW() + orderinc
    if (fe1.Space() == mfem.FunctionSpace.rQk):
        assert False, "not supported"
    ir = mfem.IntRules.Get(fe1.GetGeomType(), order)
    
    if verbose:
        dprint1("Order, N Points", order, ir.GetNPoints())
    return ir

def profile_start():
    '''
    profiling start
    usage:
       pr = profile_start()
       ... do something
       profile_stop(pr)
    '''
    import cProfile
    print('starting profiler')
    pr = cProfile.Profile()
    pr.enable()
    return pr


def profile_stop(pr, sortby='cumulative'):
    '''
    profile_stop(pr, sortby='cumulative'):

    end profile
    sortby = 'cumulative', 'calls', 'cumtime', 
             'file', 'filename', 'module',
             'ncalls', pcalls', 'line', 'name',
             'nfl', stdname', 'time', 'tottime'
    '''
    from six import StringIO
    import pstats
    pr.disable()
    # print 'stopped profiler'
    lsortby = ['cumulative', 'calls', 'cumtime',
               'file', 'filename', 'module',
               'ncalls', 'pcalls', 'line', 'name',
               'nfl', 'stdname', 'time', 'tottime']
    if not sortby in lsortby:
        print('invalid sortby')
        print(lsortby)
        return

    s = StringIO()
    sortby = sortby
    ps = pstats.Stats(pr, stream=s).sort_stats(sortby)
    ps.print_stats()
    print((s.getvalue()))

def convolve1d(fes1, fes2, kernel=delta, support=None,
               orderinc=5, is_complex=False,
               trial_domain='all',
               test_domain='all',
               verbose=False, coeff=None):
    '''
    fill linear operator for convolution
    \int phi_test(x) func(x-x') phi_trial(x') dx
    '''
    mat, rstart = get_empty_map(fes2, fes1, is_complex=is_complex)

    eltrans1 = fes1.GetElementTransformation(0)
    ir = get_rule(fes1.GetFE(0), fes2.GetFE(0), eltrans1, orderinc, verbose)

    shape1 = mfem.Vector()
    shape2 = mfem.Vector()

    #nicePrint("shape", mat.shape, fes2.GetNE(), fes1.GetNE())

    # communication strategy
    #   (1) x2 (ir points on test space) is collected in each nodes
    #   (2) x2 is send to other nodes
    #   (3) each nodes compute \int f(x2-x1) phi(x1)
    #   (4) non-zero results of (3) and global index should be send back

    # Step (1, 2)
    if verbose:
        dprint1("Step 1,2")    
    x2_arr = []
    i2_arr = []

    ptx = mfem.DenseMatrix(ir.GetNPoints(), 1)
    
    attrs1 = fes2.GetMesh().GetAttributeArray()
    attrs2 = fes2.GetMesh().GetAttributeArray()
    
    for i in range(fes2.GetNE()): # scan test space
        if test_domain != 'all':
            if not attrs1[i] in test_domain: continue
        eltrans = fes2.GetElementTransformation(i)
        eltrans.Transform(ir, ptx)
        x2_arr.append(ptx.GetDataArray().copy())
        i2_arr.append(i)
    if len(i2_arr) > 0:
       ptx_x2 = np.vstack(x2_arr)
       i2_arr = np.hstack(i2_arr)
    else:
       ptx_x2 = np.array([[]])
       i2_arr = np.array([])

    #nicePrint("x2 shape", ptx_x2.shape)
    if USE_PARALLEL:
        ## note: we could implement more advanced alg. to reduce
        ## the amount of data exchange..
        x2_all = comm.allgather(ptx_x2)
        i2_all = comm.allgather(i2_arr)
    else:
        x2_all = [ptx_x2]
        i2_all = [i2_arr]
    #nicePrint("x2_all shape", x2_all.shape)

    if USE_PARALLEL:
        #this is global TrueDoF (offset is not subtracted)        
        P = fes1.Dof_TrueDof_Matrix()
        P = ToScipyCoo(P).tocsr()
        VDoFtoGTDoF1 = P.indices  
        P = fes2.Dof_TrueDof_Matrix()
        P = ToScipyCoo(P).tocsr()
        VDoFtoGTDoF2 = P.indices
        
    # Step 3
    if verbose:
        dprint1("Step 3")
    vdofs1_senddata = []
    elmats_senddata = []
        
    for knode1 in range(len(x2_all)):
        x2_onenode = x2_all[knode1]
        i2_onenode = i2_all[knode1]
        elmats_all = []
        vdofs1_all = []

        # collect vdofs
        for j in range(fes1.GetNE()):
            local_vdofs = fes1.GetElementVDofs(j)
            if USE_PARALLEL:
                subvdofs2 = [VDoFtoGTDoF1[i] for i in local_vdofs]
                vdofs1_all.append(subvdofs2)
            else:
                vdofs1_all.append(local_vdofs)

        for i, x2s in zip(i2_onenode, x2_onenode): # loop over fes2
            nd2 = len(x2s)
            #nicePrint(x2s)
            elmats = []
            for j in range(fes1.GetNE()):
                if trial_domain != 'all':
                    if not attrs1[j] in trial_domain: continue
                
                # collect integration
                fe1 = fes1.GetFE(j)
                nd1 = fe1.GetDof()
                shape1.SetSize(nd1)
                eltrans = fes1.GetElementTransformation(j)

                tmp_int = np.zeros(shape1.Size(), dtype=mat.dtype)
                elmat = np.zeros((nd2, nd1), dtype=mat.dtype)

                #if myid == 0: print("fes1 idx", j)

                dataset = []
                for jj in range(ir.GetNPoints()):
                    ip1 = ir.IntPoint(jj)
                    eltrans.SetIntPoint(ip1)
                    x1 = eltrans.Transform(ip1)[0]
                    fe1.CalcShape(ip1, shape1)
                    w = eltrans.Weight() * ip1.weight
                    dataset.append((x1, w, shape1.GetDataArray().copy()))
                    
                has_contribution = False
                for kkk, x2 in enumerate(x2s):
                    tmp_int *= 0.0
                    
                    for x1, w, shape_arr in dataset:
                        if support is not None:
                            s = support((x1 + x2)/2.0)
                            if np.abs(x1-x2) > s:
                                continue

                        has_contribution = True
                        #if myid == 0: print("check here", x1, x2)
                        val = kernel(x2-x1, (x2+x1)/2.0, w=w)
                        if coeff is not None:
                            val = val* coeff((x2+x1)/2.0)

                        #shape_arr *= w*val
                        tmp_int += shape_arr*w*val
                    elmat[kkk, :] = tmp_int

                if has_contribution:
                    elmats.append((j, elmat))
                #print(elmats)
            if len(elmats) > 0:
                elmats_all.append((i, elmats))
                
        vdofs1_senddata.append(vdofs1_all)
        elmats_senddata.append(elmats_all)
        
        # send this information to knodes;
        '''
        if USE_PARALLEL:
            #nicePrint(vdofs1_all)
            #nicePrint("elmats", [len(x) for x in elmats_all])
            if myid == knode1:
                vdofs1_data = comm.gather(vdofs1_all, root=knode1)
                elmats_data = comm.gather(elmats_all, root=knode1)
            else:
                _ = comm.gather(vdofs1_all, root=knode1)
                _ = comm.gather(elmats_all, root=knode1)
        else:
            vdofs1_data = [vdofs1_all,]
            elmats_data = [elmats_all,]
        '''
    if USE_PARALLEL:
        knode1 = 0
        for vdofs1_all, elmats_all in zip(vdofs1_senddata, elmats_senddata):
            if myid == knode1:
                vdofs1_data = comm.gather(vdofs1_all, root=knode1)
                elmats_data = comm.gather(elmats_all, root=knode1)
            else:
                _ = comm.gather(vdofs1_all, root=knode1)
                _ = comm.gather(elmats_all, root=knode1)
            knode1 = knode1 + 1
    else:
        vdofs1_data = vdofs1_senddata
        elmats_data = elmats_senddata
        
    # Step 4
    if verbose:
        dprint1("Step 4")        
    shared_data = []
    mpi_rank = 0
    for vdofs1, elmats_all in zip(vdofs1_data, elmats_data): # loop over MPI nodes
        #nicePrint("len elmats", len(elmats_all))
        #for i, elmats in enumerate(elmats_all):  # corresponds to loop over fes2
        
        if verbose:
            coupling = [len(elmats) for i, elmats in elmats_all]
            nicePrint("Element coupling for rank", mpi_rank)
            nicePrint("   Average :", (0 if len(coupling)==0 else np.mean(coupling)))
            nicePrint("   Max/Min :", (0 if len(coupling)==0 else np.max(coupling)),
                                      (0 if len(coupling)==0 else np.min(coupling)))
            mpi_rank += 1
                
        for i, elmats in elmats_all:  # corresponds to loop over fes2
            vdofs2 = fes2.GetElementVDofs(i)
            fe2 = fes2.GetFE(i)
            nd2 = fe2.GetDof()
            shape2.SetSize(nd2)

            eltrans = fes2.GetElementTransformation(i)

            #for j, elmat in enumerate(elmats):                
            for j, elmat in elmats:
                #print(vdofs1[j], elmat.shape)
                #if elmat is None:
                #    continue
                
                mm = np.zeros((len(vdofs2), len(vdofs1[j])), dtype=float)

                for ii in range(ir.GetNPoints()):
                    ip2 = ir.IntPoint(ii)
                    eltrans.SetIntPoint(ip2)
                    ww = eltrans.Weight() * ip2.weight
                    fe2.CalcShape(ip2, shape2)
                    shape2 *= ww

                    tmp_int = elmat[ii, :]
                    tmp = np.dot(np.atleast_2d(shape2.GetDataArray()).transpose(),
                                 np.atleast_2d(tmp_int))
                    mm = mm + tmp
                    #print("check here", myid, mm.shape, tmp.shape)

                # merge contribution to final mat
                if USE_PARALLEL:
                    vdofs22 = [fes2.GetLocalTDofNumber(ii) for ii in vdofs2]
                    vdofs22g = [VDoFtoGTDoF2[ii] for ii in vdofs2]
                    kkk = 0
                    for v2, v2g in zip(vdofs22, vdofs22g):
                        if v2 < 0:
                            shared_data.append([v2g, mm[kkk, :], vdofs1[j]])
                        kkk = kkk + 1

                for k, vv in enumerate(vdofs1[j]):
                    try:
                        if USE_PARALLEL:
                            mmm = mm[np.where(np.array(vdofs22) >= 0)[0], :]                            
                            vdofs222 = [x for x in vdofs22 if x >= 0]
                        else:
                            vdofs222 = vdofs2
                            mmm = mm
                        #if myid == 1:
                        #    print("check here", vdofs2, vdofs22, vdofs222)
                        #print(mmm[:, [k]])
                        tmp = mat[vdofs222, vv] + mmm[:, [k]]
                        mat[vdofs222, vv] = tmp.flatten()
                    except:
                        import traceback
                        print("error", myid)
                        #print(vdofs1, vdofs22, vdofs222, mmm.shape, k)
                        traceback.print_exc()

    if USE_PARALLEL:
        for source_id in range(nprc):
            data = comm.bcast(shared_data, root=source_id)
            myoffset = fes2.GetMyTDofOffset()
            for v2g, elmat, vdofs1 in data:
                if v2g >= myoffset and v2g < myoffset + mat.shape[0]:
                    i = v2g - myoffset
                    #print("procesising this", myid, i, v2g, elmat, vdofs1)                
                    mat[i, vdofs1] = mat[i, vdofs1] + elmat

    from scipy.sparse import coo_matrix, csr_matrix

    if USE_PARALLEL:
        if is_complex:
            m1 = csr_matrix(mat.real, dtype=float)
            m2 = csr_matrix(mat.imag, dtype=float) 
        else:
            m1 = csr_matrix(mat.real, dtype=float)
            m2 = None
        from mfem.common.chypre import CHypreMat
        start_col = fes1.GetMyTDofOffset()
        end_col = fes1.GetMyTDofOffset() + fes1.GetTrueVSize()
        col_starts = [start_col, end_col, mat.shape[1]]
        M = CHypreMat(m1, m2, col_starts=col_starts)
        #print("mat", M)
    else:
        from petram.helper.block_matrix import convert_to_ScipyCoo

        M = convert_to_ScipyCoo(coo_matrix(mat, dtype=mat.dtype))
                
    return M


def convolve2d(fes1, fes2, kernel=delta, support=None,
               orderinc=5, is_complex=False,
               trial_domain='all',
               test_domain='all',
               verbose=False, coeff=None):
    '''
    fill linear operator for convolution
    \int phi_test(x) func(x-x') phi_trial(x') dx
    
    Genralized version to multi-dim
    test/trial
        ScalarFE, ScalarFE   : func is scalar
        VectorFE, ScalarFE   : func is vector (vertical)
        ScalarFE, VectorFE   : func is vector (horizontal)
        VectorFE, VectorFE   : func matrix
    '''
    mat, rstart = get_empty_map(fes2, fes1, is_complex=is_complex)

    if fes1.GetNE() == 0:
        assert False, "FESpace does not have element"
    eltrans1 = fes1.GetElementTransformation(0)
    ir = get_rule(fes1.GetFE(0), fes2.GetFE(0), eltrans1, orderinc, verbose)

    name_fes1 = fes1.FEColl().Name()[:2]
    name_fes2 = fes2.FEColl().Name()[:2]


    sdim = fes1.GetMesh().SpaceDimension()
    if name_fes1 in ['RT', 'ND']:
        shape1 = mfem.DenseMatrix()
        vdim1 = fes1.GetMesh().SpaceDimension()
    else:
        shape1 = mfem.Vector()
        vdim1 = 1
    if name_fes2 in ['RT', 'ND']:
        shape2 = mfem.DenseMatrix()
        vdim2 = fes1.GetMesh().SpaceDimension()        
    else:
        shape2 = mfem.Vector()
        vdim1 = 1        
    
    #nicePrint("shape", mat.shape, fes2.GetNE(), fes1.GetNE())

    # communication strategy
    #   (1) x2 (ir points on test space) is collected in each nodes
    #   (2) x2 is send to other nodes
    #   (3) each nodes compute \int f(x2-x1) phi(x1)
    #   (4) non-zero results of (3) and global index should be send back

    # Step (1, 2)
    if verbose:
        dprint1("Step 1,2")    
    x2_arr = []
    i2_arr = []

    ptx = mfem.DenseMatrix(ir.GetNPoints(), sdim)
    
    attrs1 = fes2.GetMesh().GetAttributeArray()
    attrs2 = fes2.GetMesh().GetAttributeArray()

    for i in range(fes2.GetNE()): # scan test space
        if test_domain != 'all':
            if not attrs1[i] in test_domain: continue
        eltrans = fes2.GetElementTransformation(i)
        eltrans.Transform(ir, ptx)
        x2_arr.append(ptx.GetDataArray().copy().transpose())
        i2_arr.append(i)
        
    if support is not None:
        supports = np.array([support(np.mean(xxx, 0))  for xxx in  x2_arr])
    else:
        supports = -np.ones(len(x2_arr))

    if len(i2_arr) > 0:
       ptx_x2 = np.stack(x2_arr)
       i2_arr = np.hstack(i2_arr)
    else:
       ptx_x2 = np.array([[[]]])
       i2_arr = np.array([])

    #nicePrint("x2 shape", ptx_x2.shape)
    if USE_PARALLEL:
        ## note: we could implement more advanced alg. to reduce
        ## the amount of data exchange..
        x2_all = comm.allgather(ptx_x2)
        i2_all = comm.allgather(i2_arr)
        s_all = comm.allgather(supports)
    else:
        x2_all = [ptx_x2]
        i2_all = [i2_arr]
        s_all = [supports]
    #nicePrint("x2_all shape", supports.shape, len(x2_all), [tmp.shape for tmp in x2_all])

    if USE_PARALLEL:
        #this is global TrueDoF (offset is not subtracted)        
        P = fes1.Dof_TrueDof_Matrix()
        P = ToScipyCoo(P).tocsr()
        VDoFtoGTDoF1 = P.indices  
        P = fes2.Dof_TrueDof_Matrix()
        P = ToScipyCoo(P).tocsr()
        VDoFtoGTDoF2 = P.indices
        
    # Step 3
    if verbose:
        dprint1("Step 3")
    vdofs1_senddata = []
    elmats_senddata = []
        
    for knode1 in range(len(x2_all)):
        #dprint1("new knode1", myid, knode1)
            
        x2_onenode = x2_all[knode1]
        i2_onenode = i2_all[knode1]
        s_onenode = s_all[knode1]
        
        elmats_all = []
        vdofs1_all = []

        # collect vdofs
        for j in range(fes1.GetNE()):
            local_vdofs = fes1.GetElementVDofs(j)
            local_vdofs = [vv if vv >=0 else  -1 -vv for vv in local_vdofs]
            if USE_PARALLEL:
                subvdofs2 = [VDoFtoGTDoF1[i] for i in local_vdofs]
                vdofs1_all.append(subvdofs2)
            else:
                vdofs1_all.append(local_vdofs)

        #if myid == 0:
        #    pr = profile_start()

        for i, x2s, su in zip(i2_onenode, x2_onenode, s_onenode): # loop over fes2
            nd2 = len(x2s)
            #nicePrint("x2s", i, x2s.shape, x2s)
            elmats = []
            for j in range(fes1.GetNE()):
                
                if trial_domain != 'all':
                    if not attrs1[j] in trial_domain: continue

                # collect integration
                fe1 = fes1.GetFE(j)
                nd1 = fe1.GetDof()
                eltrans = fes1.GetElementTransformation(j)
                dof_sign1 = np.array([1 if vv >=0 else -1
                             for vv in fes1.GetElementVDofs(j)])

                if name_fes1 in ['RT', 'ND']:
                    shape1.SetSize(nd1, vdim1)
                else:
                    shape1.SetSize(nd1)
                elmat = np.zeros((nd2, vdim2, nd1), dtype=mat.dtype)
                tmp_int = np.zeros((vdim2, nd1), dtype=mat.dtype).squeeze()

                #if myid == 0: print("fes1 idx", j)

                dataset = []
                shapes = []                
                for jj in range(ir.GetNPoints()):
                    ip1 = ir.IntPoint(jj)
                    eltrans.SetIntPoint(ip1)
                    x1 = eltrans.Transform(ip1)
                    if name_fes1 in ['RT', 'ND']:
                        fe1.CalcVShape(eltrans, shape1)
                    else:
                        fe1.CalcShape(ip1, shape1)
                    w = eltrans.Weight() * ip1.weight
                    ss = shape1.GetDataArray().copy()
                    
                    if len(ss.shape) > 1:
                        #dof_sign1 = dof_sign1.reshape(-1, 1)
                        ss = np.transpose(ss)
                    ss = ss * dof_sign1
                    dataset.append((x1, w, ss))
                    
                has_contribution = False
                for kkk, x2 in enumerate(x2s):
                    tmp_int *= 0.0
                    has_contribution2 = False                    
                    for x1, w, shape_arr in dataset:
                        s = np.sqrt(np.sum((x1-x2)**2))
                        if su >= 0 and s > su:
                            continue
                                          
                        val = kernel(x2-x1, (x2+x1)/2.0, w=w)
                        if val is None:
                            continue
                        if coeff is not  None:
                            val = val* coeff((x2+x1)/2.0)

                        tmp_int += np.dot(val, shape_arr)*w
                        has_contribution2 = True
                        
                    if has_contribution2:
                        elmat[kkk, ...] = tmp_int
                        has_contribution = True
                if has_contribution:
                    elmats.append((j, elmat))
                    
            #if myid == 0:
            #    pr.dump_stats("/home/shiraiwa/test.prf")                
            #    profile_stop(pr)
            #    assert False, "hoge"
            #    pr = profile_start()
            if len(elmats) > 0:
                elmats_all.append((i, elmats))
                
        vdofs1_senddata.append(vdofs1_all)
        elmats_senddata.append(elmats_all)
        
        # send this information to knodes;
        '''
        if USE_PARALLEL:
            #nicePrint(vdofs1_all)
            #nicePrint("elmats", [len(x) for x in elmats_all])
            if myid == knode1:
                vdofs1_data = comm.gather(vdofs1_all, root=knode1)
                elmats_data = comm.gather(elmats_all, root=knode1)
            else:
                _ = comm.gather(vdofs1_all, root=knode1)
                _ = comm.gather(elmats_all, root=knode1)
        else:
            vdofs1_data = [vdofs1_all,]
            elmats_data = [elmats_all,]
        '''
    if USE_PARALLEL:
        knode1 = 0
        for vdofs1_all, elmats_all in zip(vdofs1_senddata, elmats_senddata):
            if myid == knode1:
                vdofs1_data = comm.gather(vdofs1_all, root=knode1)
                elmats_data = comm.gather(elmats_all, root=knode1)
            else:
                _ = comm.gather(vdofs1_all, root=knode1)
                _ = comm.gather(elmats_all, root=knode1)
            knode1 = knode1 + 1
    else:
        vdofs1_data = vdofs1_senddata
        elmats_data = elmats_senddata
        
    # Step 4
    if verbose:
        dprint1("Step 4")        
    shared_data = []
    mpi_rank = 0
    for vdofs1, elmats_all in zip(vdofs1_data, elmats_data): # loop over MPI nodes
        #nicePrint("len elmats", len(elmats_all))
        #for i, elmats in enumerate(elmats_all):  # corresponds to loop over fes2
        
        if verbose:
            coupling = [len(elmats) for i, elmats in elmats_all]
            nicePrint("Element coupling for rank/count", mpi_rank, len(coupling))
            nicePrint("   Average :", (0 if len(coupling)==0 else np.mean(coupling)))
            nicePrint("   Max/Min :", (0 if len(coupling)==0 else np.max(coupling)),
                                      (0 if len(coupling)==0 else np.min(coupling)))
            mpi_rank += 1

        for i, elmats in elmats_all:  # corresponds to loop over fes2
            vdofs2 = fes2.GetElementVDofs(i)
            dof_sign2 = np.array([[1 if vv >= 0 else -1
                                   for vv in vdofs2],]).transpose()
            vdofs2 = [-1-x if x < 0 else x for x in vdofs2]            
            
            fe2 = fes2.GetFE(i)
            nd2 = fe2.GetDof()
            
            if name_fes2 in ['RT', 'ND']:
                shape2.SetSize(nd2, vdim2)
            else:
                shape2.SetSize(nd2)

            eltrans = fes2.GetElementTransformation(i)

            #for j, elmat in enumerate(elmats):                
            for j, elmat in elmats:
                #print(vdofs1[j], elmat.shape)
                #if elmat is None:
                #    continue
                
                mm = np.zeros((len(vdofs2), len(vdofs1[j])), dtype=float)

                for ii in range(ir.GetNPoints()):
                    ip2 = ir.IntPoint(ii)
                    eltrans.SetIntPoint(ip2)
                    ww = eltrans.Weight() * ip2.weight
                    
                    if name_fes2 in ['RT', 'ND']:
                        fe2.CalcVShape(eltrans, shape2)
                    else:
                        fe2.CalcShape(ip2, shape2)

                    shape2 *= ww
                    ss = shape2.GetDataArray().reshape(-1, vdim2)
                    ss = ss * dof_sign2

                    tmp_int = elmat[ii, ...].reshape(vdim1, -1)
                    tmp = np.dot(ss, tmp_int)
                    mm = mm + tmp

                # preapre shared data
                if USE_PARALLEL:
                    vdofs22 = [fes2.GetLocalTDofNumber(ii) for ii in vdofs2]
                    vdofs22g = [VDoFtoGTDoF2[ii] for ii in vdofs2]
                    kkk = 0
                    #for v2, v2g in zip(vdofs22, vdofs22g):
                    for v2, v2g in zip(vdofs22, vdofs22g):
                        if v2 < 0:
                            shared_data.append([v2g, mm[kkk, :], vdofs1[j]])
                        kkk = kkk + 1
                        
                # merge contribution to final mat
                for k, vv in enumerate(vdofs1[j]):
                    try:
                        if USE_PARALLEL:
                            mmm = mm[np.where(np.array(vdofs22) >= 0)[0], :]                            
                            vdofs222 = [x for x in vdofs22 if x >= 0]
                        else:
                            vdofs222 = vdofs2
                            mmm = mm
                        #if myid == 1:
                        #    print("check here", vdofs2, vdofs22, vdofs222)
                        #print(mmm[:, [k]])
                        tmp = mat[vdofs222, vv] + mmm[:, [k]]
                        mat[vdofs222, vv] = tmp.flatten()

                    except:
                        import traceback
                        print("error", myid)
                        #print(vdofs1, vdofs22, vdofs222, mmm.shape, k)
                        traceback.print_exc()

    if USE_PARALLEL:
        for source_id in range(nprc):
            data = comm.bcast(shared_data, root=source_id)
            myoffset = fes2.GetMyTDofOffset()
            for v2g, elmat, vdofs1 in data:
                if v2g >= myoffset and v2g < myoffset + mat.shape[0]:
                    i = v2g - myoffset
                    #print("procesising this", myid, i, v2g, elmat, vdofs1)                
                    mat[i, vdofs1] = mat[i, vdofs1] + elmat

    from scipy.sparse import coo_matrix, csr_matrix

    if USE_PARALLEL:
        if is_complex:
            m1 = csr_matrix(mat.real, dtype=float)
            m2 = csr_matrix(mat.imag, dtype=float) 
        else:
            m1 = csr_matrix(mat.real, dtype=float)
            m2 = None
        from mfem.common.chypre import CHypreMat
        start_col = fes1.GetMyTDofOffset()
        end_col = fes1.GetMyTDofOffset() + fes1.GetTrueVSize()
        col_starts = [start_col, end_col, mat.shape[1]]
        M = CHypreMat(m1, m2, col_starts=col_starts)
    else:
        from petram.helper.block_matrix import convert_to_ScipyCoo

        M = convert_to_ScipyCoo(coo_matrix(mat, dtype=mat.dtype))
                
    return M




from petram.phys.vtable import VtableElement, Vtable, Vtable_mixin
from petram.namespace_mixin import NS_mixin
from .solver_model import SolverBase

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('SolveControl')
format_memory_usage = debug.format_memory_usage


class SolveControl(SolverBase):
    has_2nd_panel = False

    def get_phys(self):
        my_solve_step = self.get_solve_root()
        return my_solve_step.get_phys()

    def get_phys_range(self):
        my_solve_step = self.get_solve_root()
        return my_solve_step.get_phys_range()


data = [("max_count", VtableElement("max_count",
                                    type='int',
                                    guilabel="Max count",
                                    default=3,
                                    tip="parameter range",))]


class ForLoop(SolveControl, NS_mixin, Vtable_mixin):
    vt_loop = Vtable(data)

    def __init__(self, *args, **kwargs):
        SolveControl.__init__(self, *args, **kwargs)
        NS_mixin.__init__(self, *args, **kwargs)

    def attribute_set(self, v):
        v['phys_model'] = ''
        v['init_setting'] = ''
        v['postprocess_sol'] = ''
        v['dwc_name'] = ''
        v['use_dwc'] = False
        v['dwc_args'] = ''
        v['counter_name'] = 'loop_counter'
        self.vt_loop.attribute_set(v)

        super(ForLoop, self).attribute_set(v)
        return v

    def get_possible_child(self):
        from petram.solver.solver_model import SolveStep
        from petram.solver.parametric import Parametric
        return [SolveStep, Parametric, Break, Continue, DWCCall]

    def panel1_param(self):
        panels = self.vt_loop.panel_param(self)

        ret1 = [["dwc", self.dwc_name, 0, {}, ],
                ["args.", self.dwc_args, 0, {}, ], ]
        value1 = [self.dwc_name, self.dwc_args]
        panel2 = [[None, [False, value1, ], 27, [{'text': 'use DWC (postprocess)'},
                                                 {'elp': ret1}, ]], ]

        return ([["Postprocess solution", self.postprocess_sol, 0, {}, ],
                 ["Counter name", self.counter_name, 0, {}, ]] + panels + panel2)

    def get_panel1_value(self):
        val = self.vt_loop.get_panel_value(self)

        return (  # self.init_setting,
            self.postprocess_sol,
            self.counter_name,
            val[0],
            [self.use_dwc, [self.dwc_name, self.dwc_args]])

    def import_panel1_value(self, v):
        #self.init_setting = v[0]
        self.postprocess_sol = v[0]
        self.counter_name = v[1]
        self.vt_loop.import_panel_value(self, (v[2],))
        self.use_dwc = v[3][0]
        self.dwc_name = v[3][1][0]
        self.dwc_args = v[3][1][1]

    def get_all_phys_range(self):
        steps = self.get_active_steps()
        ret0 = sum([s.get_phys_range() for s in steps], [])

        ret = []
        for x in ret0:
            if not x in ret:
                ret.append(x)

        return ret

    def get_active_steps(self, with_control=False):
        steps = []
        for x in self.iter_enabled():
            if not x.enabled:
                continue

            if isinstance(x, Break) and with_control:
                steps.append(x)
            elif isinstance(x, Continue) and with_control:
                steps.append(x)
            elif isinstance(x, DWCCall) and with_control:
                steps.append(x)
            elif len(list(x.iter_enabled())) > 0:
                steps.append(x)

        return steps

    def get_pp_setting(self):
        names = self.postprocess_sol.split(',')
        names = [n.strip() for n in names if n.strip() != '']
        return [self.root()['PostProcess'][n] for n in names]

    def iter_active_solvers(self, with_control=False):
        steps = self.get_active_steps(with_control=False)
        for s in steps:
            for s2 in s.get_active_solvers():
                yield s2

    def get_target_phys(self):
        ret = []
        for s in self.iter_active_solvers():
            for item in s.get_target_phys():
                if item not in ret:
                    ret.append(item)
        return ret

    def get_child_solver(self):
        ret = []
        for s in self.iter_active_solvers():
            ret.extend(s.get_child_solver())
        return ret

    def get_custom_init(self):
        ret = []
        for s in self.iter_active_solvers():
            ret.extend(s.get_custom_init())
        return ret

    def get_matrix_weight(self, timestep_config):  # , timestep_weight):
        ret = []
        s1, s2, s3 = 0, 0, 0
        for s in self.iter_active_solvers():
            n1, n2, n3 = s.get_matrix_weight(timestep_config)
            s1 = max(s1, n1)
            s2 = max(s2, n2)
            s3 = max(s3, n3)
        return s1, s2, s3

    def run(self, engine, is_first=True):
        dprint1("!!!!! Entering SolveLoop : (is_first =",
                is_first, ") " + self.name() + " !!!!!")

        steps = self.get_active_steps(with_control=True)
        self.vt_loop.preprocess_params(self)
        max_count = self.vt_loop.make_value_or_expression(self)[0]

        for i in range(max_count):
            dprint1("!!!!! SolveLoop : Count = " + str(i))
            g = self._global_ns[self.counter_name] = i
            for s in steps:
                do_break = False
                do_continue = False
                if isinstance(s, Break):
                    do_break = s.run(engine, i)
                elif isinstance(s, Continue):
                    do_continue = s.run(engine, i)
                elif isinstance(s, DWCCall):
                    do_continue = s.run(engine, i)
                else:
                    is_first = s.run(engine, is_first=is_first)
                    if s.solve_error[0]:
                        dprint1(
                            "Loop failed " +
                            s.name() +
                            ":" +
                            s.solve_error[1])
                        break
                    is_first = False

                if do_break or do_continue:
                    break
            if do_break:
                break

        postprocess = self.get_pp_setting()
        engine.run_postprocess(postprocess, name=self.name())

        if self.use_dwc:
            engine.call_dwc(self.get_all_phys_range(),
                            method="postprocess",
                            callername=self.name(),
                            dwcname=self.dwc_name,
                            args=self.dwc_args)


class InnerForLoop(ForLoop):
    '''
    InnerForLoop is placed insde SolveStep
    Loop supports only standard stationary solver
    '''

    def get_active_solvers(self, with_control=False):
        steps = []
        for x in self.iter_enabled():
            if not x.enabled:
                continue

            if isinstance(x, Break) and with_control:
                steps.append(x)
            elif isinstance(x, Continue) and with_control:
                steps.append(x)
            elif isinstance(x, DWCCall) and with_control:
                steps.append(x)
            elif len(list(x.iter_enabled())) > 0:
                steps.append(x)

        return steps

    def iter_active_solvers(self):
        solvers = self.get_active_solvers(with_control=False)
        for s in solvers:
            yield s

    def free_instance(self):
        for s in self.iter_active_solvers():
            s.free_instance()

    def get_num_levels(self):
        return 1

    def create_refined_levels(self, engine, lvl):
        '''
        create refined levels and return True if it is created.
        default False (no refined level)
        '''
        return False

    def get_possible_child(self):
        #from solver.solinit_model import SolInit
        from petram.solver.std_solver_model import StdSolver
        from petram.solver.mg_solver_model import MGSolver
        from petram.solver.timedomain_solver_model import TimeDomain
        from petram.solver.set_var import SetVar
        from petram.solver.distance_solver import DistanceSolver

        try:
            from petram.solver.std_meshadapt_solver_model import StdMeshAdaptSolver
            return [StdSolver,
                    TimeDomain,
                    StdMeshAdaptSolver,
                    Break, Continue,
                    DWCCall, SetVar, ]

        except:
            return [StdSolver,
                    TimeDomain,
                    Break, Continue,
                    DWCCall, SetVar]

    def get_possible_child_menu(self):
        #from solver.solinit_model import SolInit
        from petram.solver.std_solver_model import StdSolver
        from petram.solver.mg_solver_model import MGSolver
        from petram.solver.timedomain_solver_model import TimeDomain
        from petram.solver.set_var import SetVar
        from petram.solver.distance_solver import DistanceSolver

        try:
            from petram.solver.std_meshadapt_solver_model import StdMeshAdaptSolver
            return [("", StdSolver),
                    ("", TimeDomain),
                    ("extra", StdMeshAdaptSolver),
                    ("", Break),
                    ("", Continue),
                    ("", DWCCall),
                    ("!", SetVar)]
        except:
            return [("", StdSolver),
                    ("", TimeDomain),
                    ("extra", Break),
                    ("", Continue),
                    ("", DWCCall),
                    ("!", SetVar)]


class Break(SolveControl, NS_mixin):
    def __init__(self, *args, **kwargs):
        SolveControl.__init__(self, *args, **kwargs)
        NS_mixin.__init__(self, *args, **kwargs)

    def attribute_set(self, v):
        v['break_cond'] = ''
        v['use_dwc'] = False
        v['dwc_name'] = ''
        v['dwc_args'] = ''
        return super(Break, self).attribute_set(v)

    def panel1_param(self):
        ret0 = [["Break cond.", self.break_cond, 0, {}, ], ]
        ret1 = [["dwc", self.dwc_name, 0, {}, ],
                ["args.", self.dwc_args, 0, {}, ], ]
        value0 = [self.break_cond]
        value1 = [self.dwc_name, self.dwc_args]
        return [[None, [False, value1, value0], 127, [{'text': 'Use DWC (loopcontrol)'},
                                                      {'elp': ret1},
                                                      {'elp': ret0}]], ]

    def import_panel1_value(self, v):
        self.use_dwc = v[0][0]
        self.dwc_name = v[0][1][0]
        self.dwc_args = v[0][1][1]
        self.break_cond = v[0][2][0]

    def get_panel1_value(self):
        return ([self.use_dwc, [self.dwc_name, self.dwc_args],
                 [self.break_cond], ], )

    def get_all_phys_range(self):
        return self.parent.get_all_phys_range()

    def run(self, engine, count):
        if self.use_dwc:
            return engine.call_dwc(self.get_all_phys_range(),
                                   method="loopcontrol",
                                   callername=self.name(),
                                   dwcname=self.dwc_name,
                                   args=self.dwc_args,
                                   count=count,)
        else:
            if self.break_cond in self._global_ns:
                break_func = self._global_ns[self.break_cond]
            else:
                assert False, self.break_cond + " is not defined"
            g = self._global_ns
            code = "check =" + self.break_cond + '(count)'
            ll = {'count': count}
            exec(code, g, ll)

            return ll['check']


class Continue(SolveControl, NS_mixin):
    def __init__(self, *args, **kwargs):
        SolveControl.__init__(self, *args, **kwargs)
        NS_mixin.__init__(self, *args, **kwargs)

    def attribute_set(self, v):
        v['continue_cond'] = ''
        v['use_dwc'] = False
        v['dwc_name'] = ''
        v['dwc_args'] = ''
        return super(Continue, self).attribute_set(v)

    def panel1_param(self):
        ret0 = [["Continue cond.", self.continue_cond, 0, {}, ], ]
        ret1 = [["dwc", self.dwc_name, 0, {}, ],
                ["args.", self.dwc_args, 0, {}, ], ]
        value0 = [self.continue_cond]
        value1 = [self.dwc_name, self.dwc_args]
        return [[None, [False, value1, value0], 127, [{'text': 'Use DWC (loopcontrol)'},
                                                      {'elp': ret1},
                                                      {'elp': ret0}]], ]

    def import_panel1_value(self, v):
        self.use_dwc = v[0][0]
        self.dwc_name = v[0][1][0]
        self.dwc_args = v[0][1][1]
        self.continue_cond = v[0][2][0]

    def get_panel1_value(self):
        return ([self.use_dwc, [self.dwc_name, self.dwc_args],
                 [self.continue_cond], ], )

    def get_all_phys_range(self):
        return self.parent.get_all_phys_range()

    def run(self, engine, count):
        if self.use_dwc:
            return engine.call_dwc(self.get_all_phys_range(),
                                   method="loopcontrol",
                                   callername=self.name(),
                                   dwcname=self.dwc_name,
                                   args=self.dwc_args,
                                   count=count)
        else:
            if self.continue_cond in self._global_ns:
                c_func = self._global_ns[self.continue_cond]
            else:
                assert False, self.continue_cond + " is not defined"

            g = self._global_ns
            code = "check=" + self.continue_cond + '(count)'
            ll = {'count': count}
            exec(code, g, ll)

            return ll['check']


class DWCCall(SolveControl):
    '''
    standalone DWC caller
    '''

    def __init__(self, *args, **kwargs):
        SolverBase.__init__(self, *args, **kwargs)

    def attribute_set(self, v):
        v['dwc_args'] = ''
        v['dwc_name'] = ''
        v['dwc_callname'] = 'call'
        super(DWCCall, self).attribute_set(v)
        return v

    def get_possible_child(self):
        return []

    def panel1_param(self):
        panels = [["dwc", self.dwc_name, 0, {}, ],
                  ["method", self.dwc_callname, 0, {}, ],
                  ["args.", self.dwc_args, 0, {}, ], ]

        return panels

    def get_panel1_value(self):
        return [self.dwc_name, self.dwc_callname, self.dwc_args]

    def import_panel1_value(self, v):
        self.dwc_name = v[0]
        self.dwc_callname = v[1]
        self.dwc_args = v[2]

    def get_target_phys(self):
        return []

    def get_child_solver(self):
        return []

    def get_matrix_weight(self, _timestep_config):
        return []

    def get_custom_init(self):
        return []

    def get_all_phys(self):
        phys_root = self.root()['Phys']
        return [x for x in phys_root.iter_enabled()]

    def run(self, engine, is_first=True):
        engine.call_dwc(self.get_all_phys(),
                        method=self.dwc_callname,
                        callername=self.name(),
                        dwcname=self.dwc_name,
                        args=self.dwc_args)



from collections import defaultdict
from itertools import combinations
from sets import Set
import numpy as np


def merge_domain_3d(mesh, domain_list, reorder_dom=True):

    if not hasattr(mesh, 'extended_connectivity'): 
        from petram.mesh.mesh_utils import  get_extended_connectivity       
        get_extended_connectivity(mesh)               
    ec = mesh.extended_connectivity

    v2s = ec['vol2surf']
    s2l = ec['surf2line']

    from petram.mesh.mesh_utils import vol2line, line2surf,line2vol
    
    v2l = vol2line(v2s, s2l)
    l2v = line2vol(v2l)
    l2s = line2surf(s2l)
    
    # find surfaces to remove
    # find edge to remove
    remove_surf=[]
    remove_edge = Set()
    merge_vol = {}
    for d in domain_list:
       merge_vol[d] = 0
    touched_dom = []    
    for i, j in combinations(domain_list, 2):
        iscts = np.intersect1d(v2s[i], v2s[j])
        if len(iscts) != 0: touched_dom.append((i,j))
        remove_surf.append(iscts)
    remove_surf = list(np.hstack(remove_surf).astype(int))
    
    for s in remove_surf:
        for l in s2l[s]:
            for v in v2l:
                if l in v2l[v] and not v in domain_list:
                    break
            else:
                remove_edge.add(l)
        for l in s2l[s]:
            connected_surf = [s for s in l2s[l] if not s in remove_surf]
            if len(connected_surf) != 2: continue
            v1 = [v for v in v2s if not v in domain_list and connected_surf[0] in v2s[v]]
            v2 = [v for v in v2s if not v in domain_list and connected_surf[1] in v2s[v]]
            if len(Set(v1+v2)) == 1:
               remove_edge.add(l)                        
               
    remove_edge = list(remove_edge)                


    # check merging to one domain
    if len(touched_dom)==0:
        assert False, "domains are not connected"
    c = list(touched_dom[0])
    for i in range(len(touched_dom)):
         for j in range(i, len(touched_dom)):
             if touched_dom[j][0] in c and not touched_dom[j][1] in c:
                 c.append(touched_dom[j][1])
             if touched_dom[j][1] in c and not touched_dom[j][0] in c:
                 c.append(touched_dom[j][0])
    print("touched_domain", c)
    if len(c) != len(domain_list): 
        assert False, "domains are not connected"       
    
    # collect faces to merge
    merge_face0 = []
    for k, l in enumerate(remove_edge):
        for f in merge_face0:
            if np.intersect1d(list(f), l2s[l]).size != 0:
                 for s in l2s[l]:
                     if not s in remove_surf: f.add(s)
                 break
        else:
            sss = [s for s in l2s[l] if not s in remove_surf]
            merge_face0.append(Set(sss))
    print(merge_face0)
    merge_face = {}
    for k, ss in enumerate(merge_face0):
        for s in ss:
            merge_face[s] = k
                 
    # mapping of bdr (face) attributes
    idx = 1
    bdrattr_map = {}
    for s in s2l:
        if s in remove_surf: continue
        if not s in merge_face:
            bdrattr_map[s] = idx
            idx = idx + 1
    for s, k in merge_face.items():
        if s in remove_surf: continue        
        bdrattr_map[s] = k + idx        

    # mapping of domaon attributes
    if reorder_dom:
        idx = 1
        domattr_map = {}
        for v in v2s:
            if not v in merge_vol:
                domattr_map[v] = idx
                idx = idx + 1
        for v, k in merge_vol.items():
            domattr_map[v] = k + idx
    else:
        domattr_map = {}
        for v in v2s:
            if not v in merge_vol:
                domattr_map[v] = v
        idx = max(v2s.keys())+1
        for v, k in merge_vol.tems():
            domattr_map[v] = k + idx
        
    print("merge_vol", merge_vol)
    print("merge_surf", merge_face)    
    print("remove_edge", remove_edge)
    print("remove_surf", remove_surf)

    print("domattr_map", domattr_map)
    print("bdrattr_map", bdrattr_map)    

    # check if edge can be removed...
        
    
    NV = mesh.GetNV()
    NE = mesh.GetNE()
    NBE = mesh.GetNBE()    

    bdrattr = mesh.GetBdrAttributeArray()
    domattr = mesh.GetAttributeArray()
    
    Nbelem = np.sum([x in bdrattr_map for x in bdrattr])

    print("NV, NE, NBE", NV, NE, Nbelem)

    import sys
    if 'mfem.par' in sys.modules:
       import mfem.par as mfem
    else:
       import mfem.par as mfem        
    omesh = mfem.Mesh(3, NV, NE, Nbelem, 3)

    for k in range(NV):
        v = mesh.GetVertexArray(k)
        omesh.AddVertex(list(v))

    for k in range(NE):
        iv = mesh.GetElementVertices(k)
        a = domattr_map[domattr[k]]
        if len(iv) == 4:
            omesh.AddTet(list(iv), a)
        if len(iv) == 8:
            omesh.AddHex(list(iv), a)
            
    for k in range(NBE):
        iv = mesh.GetBdrElementVertices(k)
        if not bdrattr[k] in bdrattr_map: continue
        a = bdrattr_map[bdrattr[k]]
        
        if len(iv) == 3:
            omesh.AddBdrTriangle(list(iv), a)
        if len(iv) == 4:
            omesh.AddBdrQuad(list(iv), a)

    omesh.FinalizeTopology()
    omesh.Finalize(refine=True, fix_orientation=True)
            
    return omesh    





'''

    Model Tree to stroe MFEM model parameters

'''
from petram.utils import (check_cluster_access,
                          check_addon_access)
from petram.namespace_mixin import NS_mixin
import numpy as np
from petram.model import Model

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('MFEMModel')


has_addon_access = check_addon_access()
has_cluster_access = check_cluster_access()


class MFEM_GeneralRoot(Model, NS_mixin):
    can_delete = False
    has_2nd_panel = True

    def __init__(self, *args, **kwargs):
        Model.__init__(self, *args, **kwargs)
        NS_mixin.__init__(self, *args, **kwargs)

    def get_info_str(self):
        return NS_mixin.get_info_str(self)

    def attribute_set(self, v):
        v['debug_level'] = 1
        v['dwc_object_name'] = ''
        v['mesh_gen'] = ''
        v['geom_gen'] = ''
        v['diagpolicy'] = 'one'
        v['partitioning'] = 'auto'
        v['submeshpartitioning'] = 'auto'
        v['autofilldiag'] = 'off'
        v['savegz'] = 'on'
        v['allow_fallback_nonjit'] = 'allow'
        v['debug_numba_jit'] = 'off'
        v['trim_debug_print'] = 'on'
        v['warning_control'] = 'once'
        super(MFEM_GeneralRoot, self).attribute_set(v)
        return v

    def panel1_param(self):
        txt = ["0-3: Larger number prints more info.",
               "Negative number: print from all nodes",
               "Speical debug bits",
               " 4: write essentail BC vector",
               " 8: memory check"]
        txt2 = "DWC (direct wrapper call) is for low level API access"
        return [["debug level", self.debug_level, 400, {}],
                ["", "\n".join(txt), 2, None],
                ["geom generator", self.geom_gen, 0, {}],
                ["mesh generator", self.mesh_gen, 0, {}],
                ["DWC object", "", 0, {}],
                ["", txt2, 2, None], ]

    def get_panel1_value(self):
        return (self.debug_level, None,
                self.geom_gen, self.mesh_gen, self.dwc_object_name, None)

    def import_panel1_value(self, v):
        self.debug_level = v[0]
        self.geom_gen = v[2]
        self.mesh_gen = v[3]
        self.dwc_object_name = str(v[4])
        import petram.debug
        petram.debug.debug_default_level = int(self.debug_level)

    def panel2_tabname(self):
        return "Extra."

    def panel2_param(self):
        return [["DiagPolicy", None, 1, {"values": ["one", "keep"]}],
                ["File compression", None, 1, {"values": ["on", "off"]}],
                ["Mesh partitioning", None, 1, {
                    "values": ["auto", "by attribute"]}],
                ["SubMesh partitioning", None, 1, {
                    "values": ["auto", "safe"]}],
                ["Autofill emtpy diag rows", None,
                    1, {"values": ["on", "off"]}],
                ["Fallback Python coefficient", None,
                    1, {"values": ["allow", "warn", "error", "always use Python coeff."]}],
                ["Check numba JIT process", None,
                    1, {"values": ["on", "off"]}],
                ["Trim debug print text", None,
                    1, {"values": ["on", "off"]}],
                ["Warning control", None,
                    1, {"values": ["default", "error", "ignore", "always",
                                   "module", "once"]}],
                ]

    def get_panel2_value(self):
        return (self.diagpolicy, self.savegz, self.partitioning, self.submeshpartitioning,
                self.autofilldiag, self.allow_fallback_nonjit, self.debug_numba_jit,
                self.trim_debug_print, self.warning_control)

    def import_panel2_value(self, v):
        self.diagpolicy = v[0]
        self.savegz = v[1]
        self.partitioning = v[2]
        self.submeshpartitioning = v[3]
        self.autofilldiag = v[4]
        self.allow_fallback_nonjit = v[5]
        self.debug_numba_jit = v[6]
        self.trim_debug_print = v[7]
        self.warning_control = v[8]

    def run(self):
        import petram.debug
        if petram.debug.debug_default_level == 0:
            petram.debug.debug_default_level = int(self.debug_level)

        petram.debug.trim_debug_print = bool(self.trim_debug_print == 'on')

        if not hasattr(self.root(), "_variables"):
            from petram.helper.variables import Variables
            self.root()._variables = Variables()

        if not hasattr(self, "warning_control"):
            self.warning_control = 'once'

        self.root()._parameters = {}
        self.root()._init_done = True

    def get_defualt_local_ns(self):
        '''
        GeneralRoot Namelist knows basic functions
        '''
        import petram.helper.functions
        return petram.helper.functions.f.copy()

    def save_attribute_set(self, skip_def_check):
        ret = Model.save_attribute_set(self, skip_def_check)
        return [x for x in ret if x != '_variable']


class MFEM_PhysRoot(Model):
    can_delete = False
    has_2nd_panel = False

    def get_possible_child(self):
        ans = []
        from petram.helper.phys_module_util import all_phys_models
        models, classes = all_phys_models()

        tmp = sorted([(cls.__name__, cls) for cls in classes])
        classes = [x[1] for x in tmp]

        return classes

    def get_possible_child_menu(self):
        '''
        return hierachial menus
        '''
        ans = []
        from petram.helper.phys_module_util import all_phys_models
        models, classes = all_phys_models()

        name_class = sorted([(cls.__name__, cls) for cls in classes])

        import wx
        petram_model = wx.GetApp().TopWindow.proj.setting.parameters.eval('PetraM')
        mesh = petram_model.variables.eval('mesh')
        if mesh is None:
            return []

        sdim = mesh.SpaceDimension()

        allkeys = ['3D', '2D', '2Da', '1D']
        if sdim == 3:
            keys = ['3D']
        elif sdim == 2:
            keys = ['2D', '2Da']
        else:
            keys = ['1D']

        menus = []

        for n, cls in name_class:
            hitkey = ''
            for k in allkeys:
                if n.find(k) != -1:
                    hitkey = k
                    break
            if hitkey == '':
                menus.append(("", cls))
            else:
                if hitkey in keys:
                    menus.append(("", cls))

        return menus

    def make_solvars(self, solsets, g=None):
        from petram.mesh.mesh_utils import (get_extended_connectivity,
                                            get_reverse_connectivity)
        from petram.helper.variables import Variable
        from petram.helper.variables import add_scalar

        for phys in self.iter_enabled():
            for mm in phys.walk_enabled():
                mm.compile_coeffs()

        solvars = [None] * len(solsets)
        if g is None:
            g = {}
        for k, v in enumerate(solsets):
            mesh, soldict = v

            get_extended_connectivity(mesh[0])
            get_reverse_connectivity(mesh[0])

            solvar = g.copy()
            for phys in self.iter_enabled():
                phys.soldict_to_solvars(soldict, solvar)

            for var in solvar.values():
                if isinstance(var, Variable):
                    var.add_topological_info(mesh[0])

            solvars[k] = solvar

            if 'minSJac_' + str(k) in soldict:
                add_scalar(solvar, "minSJac", "", ['x', 'y', 'z'],
                           soldict['minSJac_' + str(k)][0], None)
        # print "solvars", solvars
        return solvars

    def onItemSelChanged(self, evt):
        '''
        GUI response when model object is selected in
        the dlg_edit_model
        '''
        viewer = evt.GetEventObject().GetTopLevelParent().GetParent()
        viewer.set_view_mode('phys')

    def dependent_values(self, include_disabled=False):
        '''
        return dependent_values
           names: name of values
           pnames: list of physics module
           pindex: index of dependent value in the physics module
        '''
        if include_disabled:
            method = self.get_children
        else:
            method = self.iter_enabled

        names = sum([c.dep_vars for c in method()], [])
        pnames = sum([[c.name()] * len(c.dep_vars)
                      for c in method()], [])
        pindex = sum([list(range(len(c.dep_vars)))
                      for c in method()], [])

        return names, pnames, pindex

    '''
    def get_num_matrix(self, get_matrix_weight, phys_target = None):
        # get_matrix_weight: solver method to evaulate matrix weight
        if phys_target is None:
             phys_target = [self[k] for k in self]

        num_matrix = 0
        for phys in phys_target:
            for mm in phys.walk():
                if not mm.enabled: continue
                mm.set_matrix_weight(get_matrix_weight)

                wt = np.array(mm.get_matrix_weight())
                tmp = int(np.max((wt != 0)*(np.arange(len(wt))+1)))
                num_matrix = max(tmp, num_matrix)
        return num_matrix
    '''

    def get_num_matrix(self, get_matrix_weight, phys_target=None):
        import traceback
        traceback.print_stack()
        assert False, "this should not be called"

    def all_dependent_vars(self, num_matrix, phys_target, phys_range):
        '''
        FES variable + extra variable
        '''
        dep_vars = []
        isFesvars_g = []

        phys_target = phys_target if phys_target is not None else [
            self[k] for k in self]

        for phys in phys_target:
            # if not phys.enabled: continue
            dv = phys.dep_vars
            dep_vars.extend(dv)
            extra_vars = []
            for mm in phys.walk():
                if not mm.enabled:
                    continue
                for j in range(num_matrix):
                    for k in range(len(dv)):
                        for phys2 in phys_range:
                            # if not phys2.enabled: continue
                            if not mm.has_extra_DoF2(k, phys2, j):
                                continue
                            name = mm.extra_DoF_name2(k)
                            if not name in extra_vars:
                                extra_vars.append(name)

            dep_vars.extend(extra_vars)
        return dep_vars


class MFEM_PostProcessRoot(Model):
    can_delete = False
    has_2nd_panel = False

    def get_possible_child(self):
        from petram.postprocess.pp_model import PostProcess
        return [PostProcess]

    def onItemSelChanged(self, evt):
        '''
        GUI response when model object is selected in
        the dlg_edit_model
        '''
        viewer = evt.GetEventObject().GetTopLevelParent().GetParent()
        viewer.set_view_mode('phys', self)

    def is_viewmode_grouphead(self):
        return True

    def add_solvars(self, solsets, solvars):
        for k, v in enumerate(solsets):
            mesh, soldict = v
            solvar = solvars[k]
            for p in self.iter_enabled():
                for pp in p.iter_enabled():
                    pp.soldict_to_solvars(soldict, solvar)
            #solvars[k] = solvar
        return solvars


class MFEM_InitRoot(Model):
    can_delete = False
    has_2nd_panel = False

    def get_possible_child(self):
        from petram.init_model import InitSetting
        return [InitSetting]

    def onItemSelChanged(self, evt):
        '''
        GUI response when model object is selected in
        the dlg_edit_model
        '''
        viewer = evt.GetEventObject().GetTopLevelParent().GetParent()
        viewer.set_view_mode('phys', self)

    def is_viewmode_grouphead(self):
        return True


class MFEM_GeomRoot(Model):
    can_delete = False
    has_2nd_panel = False

    def get_possible_child(self):
        ret = []

        try:
            from petram.geom.occ_geom_model import OCCGeom
            ret.append(OCCGeom)
        except ImportError:
            pass

        try:
            from petram.geom.gmsh_geom_model import GmshGeom
            ret.append(GmshGeom)
        except ImportError:
            pass

        return ret


class MFEM_MeshRoot(Model):
    can_delete = False
    has_2nd_panel = False

    def get_possible_child(self):
        from petram.mesh.mesh_model import MFEMMesh, MeshGroup
        try:
            from petram.mesh.gmsh_mesh_model import GmshMesh
            return [MFEMMesh, GmshMesh]
        except BaseException:
            return [MFEMMesh]

    def onItemSelChanged(self, evt):
        '''
        GUI response when model object is selected in
        the dlg_edit_model
        '''
        viewer = evt.GetEventObject().GetTopLevelParent().GetParent()
        viewer.set_view_mode('mesh')

    @property
    def sdim(self):
        from petram.mesh.mesh_model import MeshGroup
        for obj in self.walk():
            if isinstance(obj, MeshGroup) and obj.enabled:
                return obj.sdim
        return 1


class MFEM_SolverRoot(Model):
    can_delete = False
    has_2nd_panel = False

    def get_possible_child(self):
        from petram.solver.solver_model import SolveStep
        from petram.solver.parametric import Parametric
        #from petram.solver.solve_loop import Loop
        from petram.solver.solver_controls import ForLoop, DWCCall
        return [SolveStep, Parametric, ForLoop, DWCCall]

    def get_active_solvers(self, mm=None):
        return [x for x in self.iter_enabled()]

    def onItemSelChanged(self, evt):
        '''
        GUI response when model object is selected in
        the dlg_edit_model
        '''
        viewer = evt.GetEventObject().GetTopLevelParent().GetParent()
        viewer.set_view_mode('phys', self)

    def is_viewmode_grouphead(self):
        return True

    def get_phys(self):
        phys_root = self.root()['Phys']
        ret = []
        for k in self.keys():
            for x in self[k].get_target_phys():
                if not x in ret:
                    ret.append(x)
        return ret

    def get_special_menu(self, evt):
        return [["+Run...", None, None, ],
                ["Serial", self.run_serial, None, ],
                ["Parallel", self.run_parallel, None, ],
                ["Cluster...", self.run_cluster, None, ],
                ["!", None, None, ], ]

    def run_serial(self, evt):
        evt.GetEventObject().GetParent().onSerDriver(evt)

    def run_parallel(self, evt):
        evt.GetEventObject().GetParent().onParDriver(evt)

    def run_cluster(self, evt):
        evt.GetEventObject().GetParent().onServerSolve(evt)


try:
    from petram.geom.geom_model import GeomBase
    has_geom = True
except BaseException:
    import traceback
    traceback.print_exc()
    has_geom = False


class MFEM_ModelRoot(Model):
    def __init__(self, **kwargs):
        super(MFEM_ModelRoot, self).__init__(**kwargs)
        self['General'] = MFEM_GeneralRoot()

        if has_geom:
            self['Geometry'] = MFEM_GeomRoot()
        self['Mesh'] = MFEM_MeshRoot()
        self['Phys'] = MFEM_PhysRoot()
        self['InitialValue'] = MFEM_InitRoot()
        self['PostProcess'] = MFEM_PostProcessRoot()
        self['Solver'] = MFEM_SolverRoot()

        from petram.helper.variables import Variables
        self._variables = Variables()
        self._parameters = {}

    def attribute_set(self, v):
        from petram.helper.variables import Variables
        v['_variables'] = Variables()
        v['_parameters'] = {}
        v['enabled'] = True
        v['root_path'] = ''
        v['model_path'] = ''
        return v

    def set_root_path(self, path):
        self.root_path = path

    def get_root_path(self):
        return self.root_path

    def save_setting(self, filename=''):
        fid = open(fiilename, 'w')
        for od in self.walk():
            od.write_setting(fid)
        fid.close()

    def save_to_file(self, path, meshfile_relativepath=False):
        import petram.helper.pickle_wrapper as pickle

        if meshfile_relativepath:
            for od in self.walk():
                if hasattr(od, 'use_relative_path'):
                    od.use_relative_path()
        fid = open(path, 'wb')
        pickle.dump(self, fid)
        if meshfile_relativepath:
            for od in self.walk():
                if hasattr(od, 'use_relative_path'):
                    od.restore_fullpath()
        fid.close()



'''
   create projection matrix. This is a simpler version than the one in 
   operator.py
 
   used from engine.py to create a mapping between fes.
'''
import numpy as np
from petram.mfem_config import use_parallel
if use_parallel:
    import mfem.par as mfem
    from mpi4py import MPI
    num_proc = MPI.COMM_WORLD.size
    myid = MPI.COMM_WORLD.rank
    comm = MPI.COMM_WORLD
    from petram.helper.mpi_recipes import gather_vector, allgather_vector
    from mfem.common.mpi_debug import nicePrint
else:
    import mfem.ser as mfem

import petram.debug
dprint1, dprint2, dprint3 = petram.debug.init_dprints('Projection')


def fes_mapping(element1, order1, dim1, dim2):
    '''
    gives automatic element mapping for domain-boundary coupling

    Mapping between different dim
    From(col)      Intermediate Fes                         mbf dim
    ND(3D)         2D ND                                    any (1)
                   1D L2                                     1D (2)
    ND(2D)         3D ND                                    any (1)
                   1D L2                                     1D (2)
    RT(3D)         2D L2                                     2D (2)
                   1D x
    RT(2D)         3D x
                   1D L2                                     1D (2)
    L2(3D)         x
    L2(2D)         3D RT (Not sure if we need this...)       2D (1)
                   1D  x
    L1(1D)         3D ND (Not sure if we need this...)       1D (1)     
                   2D RT (This can be either ND or RT)       1D (1)
    H1(3D/2D/1D)   H1(3D/2D/1D)                              any
    '''
    msg = ("mapping not supported from " + element1 +
           " dim: "+str(dim1) + " " + str(dim2))
    if dim1 == dim2:
        ret = element1, order1, 2

    elif element1.startswith("ND"):
        if dim1 == 3 and dim2 == 2:
            ret = "ND_FECollection", order1, 1
        elif dim1 == 2 and dim2 == 1:
            ret = "L2_FECollection", order1-1, 2
        elif dim1 == 2 and dim2 == 3:
            ret = "ND_FECollection", order1, 1
#        elif dim1 == dim2:
#            ret = "ND_FECollection", order1, 1
        else:
            assert False, msg

    elif element1.startswith("RT"):
        if dim1 == 3 and dim2 == 2:
            ret = "L2_FECollection", order1 - 1, 2
        elif dim1 == 2 and dim2 == 1:
            ret = "L2_FECollection", order1 - 1, 2
#        elif dim1 == dim2:
#            ret = "RT_FECollection", order1, 1
        else:
            assert False, msg

    elif element1.startswith("L2"):
        if dim1 == 2 and dim2 == 3:
            ret = "RT_FECollection", order1 + 1, 1
        elif dim1 == 1 and dim2 == 2:
            ret = "ND_FECollection", order1 + 1, 1
        else:
            assert False, msg

    elif element1.startswith("H1"):
        if dim1 > dim2:
            ret = "H1_FECollection", order1, 2
        else:
            ret = "H1_FECollection", order1, 1
    else:
        assert False, msg

    return ret


def find_fes_mapping(phys1, fes_idx1, phys2, fes_idx2):
    '''
    call fes_mappinng using phys and fes_idx. This is used
    when we need this in GUI.
    '''
    emesh1 = phys1.emesh_idx
    emesh2 = phys2.emesh_idx

    fec_type1 = phys1.get_fec_type(fes_idx1)
    dim1 = phys1.dim
    sdim1 = phys1.geom_dim
    order1 = phys1.fes_order(fes_idx1)

    fec_type2 = phys2.get_fec_type(fes_idx2)
    dim2 = phys2.dim
    sdim2 = phys2.geom_dim
    order2 = phys2.fes_order(fes_idx2)

    if emesh1 == emesh2:
        return fec_type1, order1

    el, order, mbfdim = fes_mapping(fec_type1, order1, dim1, dim2)
    return el, order


def simple_projection(fes1, fes2, sel_mode, tol=1e-5, old_mapping=False):
    mesh1 = fes1.GetMesh()
    mesh2 = fes1.GetMesh()
    dim1 = mesh1.Dimension()
    dim2 = mesh2.Dimension()

    projmode = ""
    if dim2 == 3:
        if sel_mode == "domain":
            projmode = "volume"
        elif sel_mode == "boundary":
            projmode = "surface"
    elif dim2 == 2:
        if sel_mode == "domain":
            projmode = "surface"
        elif sel_mode == "boundary":
            projmode = "edge"
    elif dim2 == 1:
        if sel_mode == "domain":
            projmode = "edge"
        elif sel_mode == "boundary":
            projmode = "vertex"
    assert projmode != "", "unknow projection mode"
    if sel_mode == "domain":
        if dim1 == dim2:
            idx1 = np.unique(mesh1.GetAttributeArray())
        elif dim1 == dim2+1:
            idx1 = np.unique(mesh1.GetBdrAttributeArray())
        else:
            assert False, "unsupported mode"
        idx2 = np.unique(mesh2.GetAttributeArray())
    else:
        if dim1 == dim2:
            idx1 = np.unique(mesh1.GetBdrAttributeArray())
        elif dim1 == dim2-1:
            idx1 = np.unique(mesh1.GetAttributeArray())
        idx2 = np.unique(mesh2.GetBdrAttributeArray())

    if use_parallel:
        idx1 = list(idx1)
        idx2 = list(idx2)
        idx1 = list(set(sum(comm.allgather(idx1), [])))
        idx2 = list(set(sum(comm.allgather(idx2), [])))
    idx = np.intersect1d(idx1, idx2)
    idx1 = list(idx)
    idx2 = list(idx)

    if use_parallel:
        # we may not need this?
        idx1 = list(set(sum(comm.allgather(idx1), [])))
        idx2 = list(set(sum(comm.allgather(idx2), [])))

    dprint1("projection index ", idx1, idx2)

    from petram.helper.dof_map import notrans

    sdim1 = mesh1.SpaceDimension()
    sdim2 = mesh2.SpaceDimension()

    trans1 = notrans
    trans2 = notrans

    from petram.helper.dof_map import projection_matrix as pm
    # matrix to transfer unknown from trail to test
    M, row, col = pm(idx2, idx1, fes2, [], fes2=fes1,
                     trans1=trans2, trans2=trans1,
                     mode=projmode, tol=tol, filldiag=False,
                     old_mapping=old_mapping)
    return M



from __future__ import print_function
from petram.mfem_config import use_parallel

import traceback
import numpy as np
from scipy.sparse import csr_matrix, coo_matrix, lil_matrix

from petram.model import Pair, Bdry
from petram.phys.phys_model import Phys

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('WF_PAIRS')


if use_parallel:
    from mfem.common.parcsr_extra import ToHypreParCSR, get_row_partitioning
    import mfem.par as mfem
    from mpi4py import MPI
    num_proc = MPI.COMM_WORLD.size
    myid = MPI.COMM_WORLD.rank
    from mfem.common.mpi_debug import nicePrint
else:
    import mfem.ser as mfem
    num_proc = 1
    myid = 0

'''
   Map DoF from src surface (s1) to dst surface (s2)
'''


def make_mapper(txt, g, indvars):
    lns = {}
    tt = ['    '+indvars[j]+'=xyz['+str(j)+']' for j in range(len(indvars))]
    trans1 = ['def trans1(xyz):',
              '    import numpy as np']
    trans1.extend(tt)
    trans1.append('    return np.array(['+txt+'])')
    exec('\n'.join(trans1), g, lns)      # this defines trans1
    mapper = lns['trans1']
    return mapper


def validate_mapper(value, obj, w):
    g = obj._global_ns
    root_phys = obj.get_root_phys()
    ind_vars = [x.strip() for x in root_phys.ind_vars.split(',')]
    try:
        make_mapper(value, g, ind_vars)
        return True
    except:
        # traceback.print_exc()
        return False


class WF_PeriodicCommon(Pair, Phys):
    def __init__(self,  **kwargs):
        super(WF_PeriodicCommon, self).__init__(**kwargs)
        Phys.__init__(self)

    def attribute_set(self, v):
        v['dstmap_txt'] = "x, y"
        v['srcmap_txt'] = ""
        v['tol_txt'] = "1e-4"
        v['weight_txt'] = "1"
        v['tol'] = 1e-4
        v['map_mode'] = "surface"
        v['fes_idx'] = 0
        super(WF_PeriodicCommon, self).attribute_set(v)
        return v

    def on_created_in_tree(self):
        dim = self.get_root_phys().dim
        if dim == 3:
            self.dstmap_txt = "x, y"
        elif dim == 2:
            self.dstmap_txt = "x"
        else:
            self.dstmap_txt = ""

    def panel1_param(self):
        import wx
        dep_vars = self.get_root_phys().dep_vars

        return [["constriained FEspace", dep_vars[0], 4,
                 {"style": wx.CB_READONLY, "choices": dep_vars}],
                ['dst mapping ',  self.dstmap_txt, 0, {'validator': validate_mapper,
                                                       'validator_param': self}],
                ['src mapping ',  self.srcmap_txt, 0,  {'validator': validate_mapper,
                                                        'validator_param': self}],
                self.make_phys_param_panel('weight',  self.weight_txt),
                self.make_phys_param_panel('tol.',  self.tol_txt),
                ["", "u_dst = u_src", 2, None], ]
#                ["use Lagrange multiplier",   self.use_multiplier,  3, {"text":""}],]

    def get_panel1_value(self):
        dep_vars = self.get_root_phys().dep_vars
        txt = ", ".join(
            [x+"_dst = "+x+"_src" for x in self.get_root_phys().dep_vars])

        return (dep_vars[self.fes_idx],
                self.dstmap_txt, self.srcmap_txt, self.weight_txt,
                self.tol_txt, txt)

    def import_panel1_value(self, v):
        dep_vars = self.get_root_phys().dep_vars
        self.fes_idx = dep_vars.index(str(v[0]))
        self.dstmap_txt = str(v[1])
        self.srcmap_txt = str(v[2])
        self.weight_txt = str(v[3])
        self.tol_txt = str(v[4])

    def make_mapper(self):
        g = self._global_ns
        root_phys = self.get_root_phys()
        ind_vars = [x.strip() for x in root_phys.ind_vars.split(',')]

        dst_mapper = make_mapper(self.dstmap_txt, g, ind_vars)
        if self.srcmap_txt != '':
            src_mapper = make_mapper(self.srcmap_txt, g, ind_vars)
        else:
            src_mapper = dst_mapper
        return dst_mapper, src_mapper

    def eval_wtol(self):
        g = self._global_ns
        weight = eval(self.weight_txt, g, {})
        tol = eval(self.tol_txt, g, {})
        return weight, tol

    def preprocess_params(self, engine):
        # find normal (outward) vector...
        try:
            dst_mapper, src_mapper = self.make_mapper()
        except:
            import traceback
            traceback.print_exc()
            raise ValueError("Cannot complie mappling rule")

        try:
            weight, tol = self.eval_wtol()
        except:
            import traceback
            traceback.print_exc()
            raise ValueError(
                "Cannot evaluate dphase/tolerance to float number")

    def has_extra_DoF(self, kfes=0):
        return False

    def has_interpolation_contribution(self, kfes=0):
        return kfes == self.fes_idx

    def add_interpolation_contribution(self, engine, ess_tdof=None, kfes=0):
        dprint1("Add interpolation contribution(real)" + str(self._sel_index))
        dprint1("kfes = ", str(kfes))

        weight, tol = self.eval_wtol()
        dprint1("weight = ", str(weight))
        dprint1("tol = ", str(tol))

        dst_mapper, src_mapper = self.make_mapper()

        tdof = [] if ess_tdof is None else ess_tdof
        src = self._src_index
        dst = [x for x in self._sel_index if not x in self._src_index]

        fes = engine.get_fes(self.get_root_phys(), kfes)

        if fes.GetMesh().Dimension() == 3:
            mode = 'surface'
        elif fes.GetMesh().Dimension() == 2:
            mode = 'edge'
        elif fes.GetMesh().Dimension() == 1:
            mode = 'point'
        else:
            assert False, "Unknown coupling mode"

        from petram.helper.dof_map import projection_matrix
        M, r, c = projection_matrix(src, dst, fes, ess_tdof, fes2=fes,
                                    trans1=dst_mapper, trans2=src_mapper,
                                    weight=-weight, tol=self.tol, mode=mode)

        return M, r, c


class WF_PeriodicBdr(WF_PeriodicCommon, Bdry):
    def __init__(self,  **kwargs):
        super(WF_PeriodicBdr, self).__init__(**kwargs)
        Bdry.__init__(self)



from mfem.common.mpi_debug import nicePrint
from petram.phys.vtable import VtableElement, Vtable
from petram.mfem_config import use_parallel
import numpy as np

from petram.model import Domain, Bdry, Edge, Point, Pair
from petram.phys.coefficient import SCoeff, VCoeff
from petram.phys.phys_model import Phys, PhysModule

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('DistanceWall')

if use_parallel:
    import mfem.par as mfem
else:
    import mfem.ser as mfem


class Distance_Wall(Bdry, Phys):
    has_essential = True
    nlterms = []
    can_timedpendent = True
    has_3rd_panel = True

    def __init__(self, **kwargs):
        super(Distance_Wall, self).__init__(**kwargs)

    def attribute_set(self, v):
        Bdry.attribute_set(self, v)
        Phys.attribute_set(self, v)
        v['sel_readonly'] = False
        v['sel_index'] = []
        return v

    @property
    def vt(self):
        root_phys = self.get_root_phys()
        if not isinstance(root_phys, PhysModule):
            data = (('esse_value', VtableElement(None, type="array",
                                                 guilabel="dummy",
                                                 default="0.0",
                                                 tip="dummy")),)
            return Vtable(data)

        dep_vars = self.get_root_phys().dep_vars
        if not hasattr(self, '_dep_var_bk'):
            self._dep_var_bk = ""

        if self._dep_var_bk != dep_vars:
            dep_var = dep_vars[0]
            data = (('esse_value', VtableElement("esse_value", type="array",
                                                 guilabel=dep_var,
                                                 default="0.0",
                                                 tip="Reference for disance measurement")),)
            vt = Vtable(data)
            self._vt1 = vt
            self._dep_var_bk = dep_vars
            self.update_attribute_set()
        else:
            vt = self._vt1
        return vt

    def get_essential_idx(self, kfes):
        if kfes == 0:
            return self._sel_index
        else:
            return []

    def apply_essential(self, engine, gf, real=False, kfes=0):
        if kfes > 0:
            return
        c0 = self.vt.make_value_or_expression(self)[0]

        if real:
            dprint1("Apply Ess.(real)" + str(self._sel_index), 'c0', c0)
        else:
            dprint1("Apply Ess.(imag)" + str(self._sel_index), 'c0', c0)

        name = self.get_root_phys().dep_vars[0]
        fes = engine.get_fes(self.get_root_phys(), name=name)

        fec_name = fes.FEColl().Name()

        mesh = engine.get_emesh(mm=self)
        ibdr = mesh.bdr_attributes.ToList()

        bdr_attr = [0]*mesh.bdr_attributes.Max()
        for idx in self._sel_index:
            bdr_attr[idx-1] = 1

        ess_vdofs = mfem.intArray()
        fes.GetEssentialVDofs(mfem.intArray(bdr_attr), ess_vdofs, 0)
        vdofs = np.where(np.array(ess_vdofs.ToList()) == -1)[0]
        dofs = mfem.intArray([fes.VDofToDof(i) for i in vdofs])
        fes.BuildDofToArrays()

        coeff1 = SCoeff(c0, self.get_root_phys().ind_vars,
                        self._local_ns, self._global_ns,
                        real=real)
        gf.ProjectCoefficient(coeff1, dofs, 0)
